// A Java version for the Kaser Interpreter - the Exec loop
// For some of my own fun!
// Stephen M. Bennett
// 29Jun02 Started

package stephen.bennett.kint;


public class ExecKE {
  // indicates if signed maths used
  private static boolean signedmath = false;

  // result to hold MSbit, EQ, LSbit
  private static int result = 0;     


  // for debugging - which ops are used
  // private static boolean[] usedOps = new boolean[0x100];
  // private static boolean[] usedFEOps = new boolean[0x22];


  // print out which op codes have not been exercised
  // public static void printUnusedOps() {
  //   System.out.println("Here are the opcodes that have not been exercised");
  //   for(int i=0; i<usedOps.length; i++) {
  //     if (!usedOps[i]) {
  //       System.out.println("  "+Integer.toHexString(i));
  //     }
  //   }
  //   System.out.println("End Of List");
  //
  //   System.out.println("Here are the FE opcodes that have not been exercised");
  //   for(int i=0; i<usedFEOps.length; i++) {
  //     if (!usedFEOps[i]) {
  //       System.out.println("  "+Integer.toHexString(i));
  //     }
  //   }
  //   System.out.println("End Of List");
  // }



  // run the interpreter given registers and code
//************************************************
//* The K interpreter
//*             Event Kode must put return value into R23 before HALTing
//*
//* NOTE: "result" contains the KCPU "flags" for equality,
//*             least-significant bit, and most-significant bit.
//*             The "result" flags are set by these instructions:
//*                     SHL SHR ROL ROR SUB B2I "RET 0" and "RET 1"
//*                     AND XOR CMP INC DEC NOT NEG TST
//*             When the operation is a WORD or BYTE, the result
//*             is first cast to a SHORT or a CHAR before being
//*             cast to a DWORD in order to make sure that the
//*             most-significant bit gets propagated correctly.
//
// Note for Java: result = b  or result = s (where b is byte and s is short)
//                will maintain MSB
//
//*
//*             The instructions that are affected by the 'signedmath' flag are:
//*                     SHR, DIV, MOD, SUB, CMP
  public static int execKE(byte[] regs, byte[] kode) {
    // print the registers before we start
    Log.registers(regs);

    // program counter (or ptr code) - start at value of reg3
    // and it is pc into kode
    // remember pc+1 moves a byte, pc+2 a word and pc+4 a dword
    int pc = Mem.GETREGD(regs, Mem.REG_3);
    Log.fine("pc", pc);

    // often used vars - not a clean way of programming but it means I do not
    // need a  {} for each of the case statements
    int reg = 0;
    int reg2 = 0;
    int offset = 0;
    byte bVal = 0;
    short wVal = 0;
    int dVal = 0;
    int count = 0;
    int jmpoffset = 0;
    int newpc = 0;

    // do until halt required
    boolean haltRequired = false;
    while (!haltRequired) {
      // let's see result's current value
      Log.fine("at start of next opcode result", result);
      Log.fine("pc", pc);

      // get the next opcode
      int opcode = Mem.GETMEMW(kode, pc) & 0xffff;
      Log.fine("opcode", opcode);

      // move the pc on to the next word
      pc = pc+2;

      // top byte of opcode decides on the functionality to use
      int opcodeinst = (opcode>>>8) & 0xff;

      // debugging which opcodes are used
      // if (!usedOps[opcodeinst]) {
      //   System.out.println("opcode "+Integer.toHexString(opcodeinst)+" NOT USED BEFORE");
      //  usedOps[opcodeinst] = true;
      // }



      // most take rest of opcode as reg, so get here
      reg = opcode & 0xff;
      Log.fine("reg", reg);

      switch(opcodeinst) {

        //********************************************************
        // 2-reg register ops
        //********************************************************


      case 0x03: // SHL b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg)<<(byte)(Mem.GETREGB(regs, reg2) & 0x0007)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x04: // SHL w,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg)<<(byte)(Mem.GETREGB(regs, reg2) & 0x000F)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x05: // SHL d,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg)<<(byte)(Mem.GETREGB(regs, reg2) & 0x001F));
        break;


      case 0x06: // SHR b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg)>>>(byte)(Mem.GETREGB(regs, reg2) & 0x0007)));
        }
        else {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg))>>>(byte)(Mem.GETREGB(regs, reg2) & 0x0007)));
        }
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x07: // SHR w,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg)>>>(byte)(Mem.GETREGB(regs, reg2) & 0x000F)));
        }
        else {
          result = Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg))>>>(byte)(Mem.GETREGB(regs, reg2) & 0x000F)));
        }
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x08: // SHR d,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg)>>>(byte)(Mem.GETREGB(regs, reg2) & 0x001F));
        }
        else {
          result = Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg))>>>(byte)(Mem.GETREGB(regs, reg2) & 0x001F)));
        }
        break;


      case 0x09: // ROL b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        count = (Mem.GETREGB(regs, reg2) & 0x07);
        Log.fine("count", count);

        bVal = Mem.GETREGB(regs, reg);
        Log.fine("bVal", bVal);

        while (count-- > 0) {
          if ((bVal & 0x80) != 0) {
            bVal = (byte)((bVal << 1) | 1);
          }
          else {
            bVal = (byte)(bVal << 1);
          }
        }
        result = Mem.SETREGB(regs, reg, bVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x0A: // ROL w,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        count = (Mem.GETREGB(regs, reg2) & 0x1F);
        Log.fine("count", count);

        wVal = Mem.GETREGW(regs, reg);
        Log.fine("wVal", wVal);

        while (count-- > 0) {
          if ((wVal & 0x8000) != 0) {
            wVal = (short)((wVal << 1) | 1);
          }
          else {
            wVal = (short)(wVal << 1);
          }
        }
        result = Mem.SETREGW(regs, reg, wVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x0B: // ROL d,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        count = (Mem.GETREGB(regs, reg2) & 0x1F);
        Log.fine("count", count);

        dVal = Mem.GETREGD(regs, reg);
        Log.fine("dVal", dVal);

        while (count-- > 0) {
          if ((dVal & 0x80000000) != 0) {
            dVal = (dVal << 1) | 1;
          }
          else {
            dVal = dVal << 1;
          }
        }
        result = Mem.SETREGD(regs, reg, dVal);
        break;


      case 0x0C: // ROR b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        count = (Mem.GETREGB(regs, reg2) & 0x07);
        Log.fine("count", count);

        bVal = Mem.GETREGB(regs, reg);
        Log.fine("bVal", bVal);

        while (count-- > 0) {
          if ((bVal & 0x01) != 0) {
            bVal = (byte)((bVal >>> 1) | 0x80);
          }
          else {
            bVal = (byte)(bVal >>> 1);
          }
        }
        result = Mem.SETREGB(regs, reg, bVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x0D: // ROR w,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        count = (Mem.GETREGB(regs, reg2) & 0x1F);
        Log.fine("count", count);

        wVal = Mem.GETREGW(regs, reg);
        Log.fine("wVal", wVal);

        while (count-- > 0) {
          if ((wVal & 0x01) != 0) {
            wVal = (short)((wVal >>> 1) | 0x8000);
          }
          else {
            wVal = (short)(wVal >>> 1);
          }
        }
        result = Mem.SETREGW(regs, reg, wVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x0E: // ROR d,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        count = (Mem.GETREGB(regs, reg2) & 0x1F);
        Log.fine("count", count);

        dVal = Mem.GETREGD(regs, reg);
        Log.fine("dVal", dVal);

        while (count-- > 0) {
          if ((dVal & 0x01) != 0) {
            dVal = (dVal >>> 1) | 0x80000000;
          }
          else {
            dVal = dVal >>> 1;
          }
        }
        result = Mem.SETREGD(regs, reg, dVal);
        break;


      case 0x0F: // ADD b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          // the java default
          Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) + Mem.GETREGB(regs, reg2)));
        }
        else {
          Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) + Mem.bUnsigned(Mem.GETREGB(regs, reg2))));
        }
        break;

      case 0x10: // ADD w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) + Mem.GETREGW(regs, reg2)));
        }
        else {
          Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) + Mem.wUnsigned(Mem.GETREGW(regs, reg2))));
        }
        break;

      case 0x11: // ADD d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) + Mem.GETREGD(regs, reg2));
        }
        else {
          Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) + Mem.dUnsigned(Mem.GETREGD(regs, reg2))));
        }
        break;


      case 0x12: // SUB b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          // the java default
          result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) - Mem.GETREGB(regs, reg2)));
        }
        else {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) - Mem.bUnsigned(Mem.GETREGB(regs, reg2))));
        }
        break;

      case 0x13: // SUB w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) - Mem.GETREGW(regs, reg2)));
        }
        else {
          result = Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) - Mem.wUnsigned(Mem.GETREGW(regs, reg2))));
        }
        break;

      case 0x14: // SUB d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) - Mem.GETREGD(regs, reg2));
        }
        else {
          result = Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) - Mem.dUnsigned(Mem.GETREGD(regs, reg2))));
        }
        break;


      case 0x15: // MUL b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          // the java default
          Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) * Mem.GETREGB(regs, reg2)));
        }
        else {
          Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) * Mem.bUnsigned(Mem.GETREGB(regs, reg2))));
        }
        break;

      case 0x16: // MUL w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) * Mem.GETREGW(regs, reg2)));
        }
        else {
          Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) * Mem.wUnsigned(Mem.GETREGW(regs, reg2))));
        }
        break;

      case 0x17: // MUL d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) * Mem.GETREGD(regs, reg2));
        }
        else {
          Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) * Mem.dUnsigned(Mem.GETREGD(regs, reg2))));
        }
        break;


      case 0x18: // DIV b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          // the java default
          Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) / Mem.GETREGB(regs, reg2)));
        }
        else {
          Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) / Mem.bUnsigned(Mem.GETREGB(regs, reg2))));
        }
        break;

      case 0x19: // DIV w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) / Mem.GETREGW(regs, reg2)));
        }
        else {
          Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) / Mem.wUnsigned(Mem.GETREGW(regs, reg2))));
        }
        break;

      case 0x1A: // DIV d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) / Mem.GETREGD(regs, reg2));
        }
        else {
          Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) / Mem.dUnsigned(Mem.GETREGD(regs, reg2))));
        }
        break;


      case 0x1B: // MOD b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          // the java default
          Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) % Mem.GETREGB(regs, reg2)));
        }
        else {
          Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) % Mem.bUnsigned(Mem.GETREGB(regs, reg2))));
        }
        break;

      case 0x1C: // MOD w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) % Mem.GETREGW(regs, reg2)));
        }
        else {
          Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) % Mem.wUnsigned(Mem.GETREGW(regs, reg2))));
        }
        break;

      case 0x1D: // MOD d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) % Mem.GETREGD(regs, reg2));
        }
        else {
          Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) % Mem.dUnsigned(Mem.GETREGD(regs, reg2))));
        }
        break;


      case 0x1E: // OR b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) | Mem.GETREGB(regs, reg2)));
        break;

      case 0x1F: // OR w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) | Mem.GETREGW(regs, reg2)));
        break;

      case 0x20: // OR d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) | Mem.GETREGD(regs, reg2));
        break;


      case 0x21: // AND b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) & Mem.GETREGB(regs, reg2)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x22: // AND w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) & Mem.GETREGW(regs, reg2)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x23: // AND d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) & Mem.GETREGD(regs, reg2));
        break;


      case 0x24: // XOR b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) ^ Mem.GETREGB(regs, reg2)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x25: // XOR w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) ^ Mem.GETREGW(regs, reg2)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x26: // XOR d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) ^ Mem.GETREGD(regs, reg2));
        break;


      case 0x27: // CMP b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          // the java default
          result = (byte)(Mem.GETREGB(regs, reg) - Mem.GETREGB(regs, reg2));
        }
        else {
          result = (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) - Mem.bUnsigned(Mem.GETREGB(regs, reg2)));
        }
        break;

      case 0x28: // CMP w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          result = (short)(Mem.GETREGW(regs, reg) - Mem.GETREGW(regs, reg2));
        }
        else {
          result = (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) - Mem.wUnsigned(Mem.GETREGW(regs, reg2)));
        }
        break;

      case 0x29: // CMP d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (signedmath) {
          result = Mem.GETREGD(regs, reg) - Mem.GETREGD(regs, reg2);
        }
        else {
          result = (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) - Mem.dUnsigned(Mem.GETREGD(regs, reg2)));
        }
        break;


      case 0x2A: // TST b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = (byte)(Mem.GETREGB(regs, reg) & Mem.GETREGB(regs, reg2));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x2B: // TST w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = (short)(Mem.GETREGW(regs, reg) & Mem.GETREGW(regs, reg2));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x2C: // TST d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        result = Mem.GETREGD(regs, reg) & Mem.GETREGD(regs, reg2);
        break;


      case 0x2D: // XCH b,b
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        bVal = Mem.GETREGB(regs, reg);
        Mem.SETREGB(regs, reg, Mem.GETREGB(regs, reg2));
        Mem.SETREGB(regs, reg2, bVal);
        break;

      case 0x2E: // XCH w,w
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        wVal = Mem.GETREGW(regs, reg);
        Mem.SETREGW(regs, reg, Mem.GETREGW(regs, reg2));
        Mem.SETREGW(regs, reg2, wVal);
        break;

      case 0x2F: // XCH d,d
        // get WORD value from memory, move pc over it
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        dVal = Mem.GETREGD(regs, reg);
        Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg2));
        Mem.SETREGD(regs, reg2, dVal);
        break;


      case 0x30: // LDR b,b
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETREGB(regs, reg2));
        break;

      case 0x31: // LDR w,w
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETREGW(regs, reg2));
        break;

      case 0x32: // LDR d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg2));
        break;


      case 0x33: // LDE d,b
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETREGB(regs, reg2) & 0xFF);
        break;

      case 0x34: // LDE d,w
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETREGW(regs, reg2) & 0xFFFF);
        break;

      case 0x35: // LDE w,b
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, (short)(Mem.GETREGB(regs, reg2) & 0xFF));
        break;

      case 0x36: // ABIT w,b
        if (reg <= Mem.REG_96 ) {
          // get reg in WORD value, moving pc over WORD
          reg2 = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg2", reg2);

          wVal = Mem.GETREGW(regs, reg);
          Log.fine("wVal", wVal);
          for(; wVal != 0; wVal = (short)((wVal&0xffff)>>>1), ++reg2) {
            if ((wVal & 1) != 0) {
              Mem.SETREGB(regs, reg2, (byte)(Mem.GETREGB(regs, reg2)+1));
            }
          }
        }
        break;

      case 0x37: // ABIT b,b
          // get reg in WORD value, moving pc over WORD
          reg2 = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg2", reg2);

          bVal = Mem.GETREGB(regs, reg);
          Log.fine("bVal", bVal);
          for(; bVal != 0; bVal = (byte)((bVal&0xff)>>>1), ++reg2) {
            if ((bVal & 1) != 0) {
              Mem.SETREGB(regs, reg2, (byte)(Mem.GETREGB(regs, reg2)+1));
            }
          }
          break;


        //***************************************************
        // 2-reg register-direct ops (2nd reg always 4 bytes)
        //***************************************************


      case 0x38: // LD0 b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, Mem.GETREGD(regs, reg2)));
        break;

      case 0x39: // LD0 w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, Mem.GETREGD(regs, reg2)));
        break;

      case 0x3A: // LD0 d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, Mem.GETREGD(regs, reg2)));
        break;


      case 0x3B: // LD1 b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2)));
        break;

      case 0x3C: // LD1 w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2)));
        break;

      case 0x3D: // LD1 d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2)));
        break;


      case 0x3E: // LD2 b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2)));
        break;

      case 0x3F: // LD2 w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2)));
        break;

      case 0x40: // LD2 d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2)));
        break;


      case 0x41: // LDS b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2)));
        break;

      case 0x42: // LDS w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2)));
        break;

      case 0x43: // LDS d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2)));
        break;


      case 0x44: // ST0 b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(kode, Mem.GETREGD(regs, reg2), Mem.GETREGB(regs, reg));
        break;

      case 0x45: // ST0 w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(kode, Mem.GETREGD(regs, reg2), Mem.GETREGW(regs, reg));
        break;

      case 0x46: // ST0 d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(kode, Mem.GETREGD(regs, reg2), Mem.GETREGD(regs, reg));
        break;


      case 0x47: // ST1 b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2), Mem.GETREGB(regs, reg));
        break;

      case 0x48: // ST1 w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2), Mem.GETREGW(regs, reg));
        break;

      case 0x49: // ST1 d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2), Mem.GETREGD(regs, reg));
        break;


      case 0x4A: // ST2 b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2), Mem.GETREGB(regs, reg));
        break;

      case 0x4B: // ST2 w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2), Mem.GETREGW(regs, reg));
        break;

      case 0x4C: // ST2 d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2), Mem.GETREGD(regs, reg));
        break;


      case 0x4D: // STS b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2), Mem.GETREGB(regs, reg));
        break;

      case 0x4E: // STS w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2), Mem.GETREGW(regs, reg));
        break;

      case 0x4F: // STS d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2), Mem.GETREGD(regs, reg));
        break;


      case 0x50: // PU0 b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(kode, Mem.GETREGD(regs, reg2), Mem.GETREGB(regs, reg));
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) + 1);
        break;

      case 0x51: // PU0 w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(kode, Mem.GETREGD(regs, reg2), Mem.GETREGW(regs, reg));
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) + 2);
        break;

      case 0x52: // PU0 d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(kode, Mem.GETREGD(regs, reg2), Mem.GETREGD(regs, reg));
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) + 4);
        break;


      case 0x53: // PO0 b,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // subtract 1 (a byte size) from this register
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) - 1);

        // set reg with the value at 2nd reg mem address
        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, Mem.GETREGD(regs, reg2)));
        break;

      case 0x54: // PO0 w,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // subtract 2 (a word size) from this register
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) - 2);

        // set reg with the value at 2nd reg mem address
        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, Mem.GETREGD(regs, reg2)));
        break;

      case 0x55: // PO0 d,d
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // subtract 4 (a dword size) from this register
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) - 4);

        // set reg with the value at 2nd reg mem address
        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, Mem.GETREGD(regs, reg2)));
        break;



        //*********************************************************
        // 2-reg register-indexed-direct ops (reg2 address 4 bytes)
        //*********************************************************

      case 0x58: // LD0 b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x59: // LD0 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x5A: // LD0 d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;


      case 0x5B: // LD1 b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x5C: // LD1 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x5D: // LD1 d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;


      case 0x5E: // LD2 b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x5F: // LD2 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x60: // LD2 d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // jump over DWORD
        pc = pc+4;
        break;


      case 0x61: // LDS b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // move pc over DWORD
        pc = pc+4;
        break;

      case 0x62: // LDS w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // move pc over DWORD
        pc = pc+4;
        break;

      case 0x63: // LDS d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        // move pc over DWORD
        pc = pc+4;
        break;


      case 0x64: // ST0 b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGB(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x65: // ST0 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGW(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x66: // ST0 d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGD(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;


      case 0x67: // ST1 b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGB(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x68: // ST1 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGW(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x69: // ST1 d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_1), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGD(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;


      case 0x6A: // ST2 b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGB(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x6B: // ST2 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGW(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x6C: // ST2 d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_2), Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGD(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;


      case 0x6D: // STS b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGB(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x6E: // STS w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGW(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;

      case 0x6F: // STS d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_4) + Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGD(regs, reg));
        // jump over DWORD
        pc = pc+4;
        break;


      case 0x70: // PU0 b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMB(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGB(regs, reg));
        pc = pc+4;
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) + 1);
        break;

      case 0x71: // PU0 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMW(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGW(regs, reg));
        pc = pc+4;
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) + 2);
        break;

      case 0x72: // PU0 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        Mem.SETMEMD(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc), Mem.GETREGD(regs, reg));
        pc = pc+4;
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) + 4);
        break;


      case 0x73: // PO0 b,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // subtract 1 (a byte size) from this register
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) - 1);

        // set reg with the value at 2nd reg mem address
        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        pc = pc+4;
        break;

      case 0x74: // PO0 w,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // subtract 2 (a word size) from this register
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) - 2);

        // set reg with the value at 2nd reg mem address
        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        pc = pc+4;
        break;

      case 0x75: // PO0 d,d,=dval
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // subtract 4 (a dword size) from this register
        Mem.SETREGD(regs, reg2, Mem.GETREGD(regs, reg2) - 4);

        // set reg with the value at 2nd reg mem address
        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, Mem.GETREGD(regs, reg2) + Mem.GETMEMD(kode, pc)));
        pc = pc+4;
        break;


      case 0x76: // LSA d,=dval
        // get reg in WORD value, moving pc over WORD
        offset = Mem.wUnsigned(Mem.GETMEMW(kode, pc));
        pc = pc+2;
        Log.fine("offset", offset);

        Mem.SETREGD(regs, reg, Mem.GETREGD(regs, Mem.REG_4) + offset);
        break;


      case 0x77: // B2I r,r (byte-2-index dst-cnt-reg,src-bit-test-reg
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        if (reg2 < 0x0080) {
          dVal = Mem.GETREGD(regs, reg2);
        }
        else if (reg2 < 0x00E0) {
          dVal = Mem.GETREGW(regs, reg2);
        }
        else {
          dVal = Mem.GETREGB(regs, reg2);
        }

        if (dVal != 0) {
          result = 0;
          while ((dVal & 1) == 0) {
            ++result;
            dVal >>>= 1;
          }
        } 
        else result = 0xFFFFFFFF;

        if (reg < 0x0080) {
          Mem.SETREGD(regs, reg, result);
        }
        else if (reg < 0x00E0) {
          Mem.SETREGW(regs, reg, (short)result);
        }
        else {
          Mem.SETREGB(regs, reg, (byte)result);
        }
        break;


        //*************************************************
        // 2-reg range ops (always byte range and inclusive
        //*************************************************

      case 0x78: // CLR b,b
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // clear the bytes for this range
        for(int i=reg; i<=reg2; i++) {
          regs[i] = 0;
        }
        break;

      case 0x79: // PUR b,b
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // copy these above the stack
        dVal = Mem.GETREGD(regs, Mem.REG_5);
        for(int i=reg; i<=reg2; i++) {
          kode[dVal++] = regs[i];
        }

        // move stack up
        Mem.SETREGD(regs, Mem.REG_5, dVal);
        break;

      case 0x7A: // POR b,b
        // get reg in WORD value, moving pc over WORD
        reg2 = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("reg2", reg2);

        // copy stack into these regs
        dVal = Mem.GETREGD(regs, Mem.REG_5);
        for(int i=reg2; i>=reg; i--) {
          regs[i] = kode[--dVal];
        }
        
        // move stack back down
        Mem.SETREGD(regs, Mem.REG_5, dVal);
        break;


        //***********************************************
        // 1-reg ops
        //***********************************************

      case 0x7D: // SAF b
        bVal = (byte)(result & 1);
        if (result == 0) bVal |= 2;
        if ((result & 0x80000000) != 0) bVal |= 4;
        if (signedmath) bVal |= 8;
        Log.fine("SAF result", bVal);
        Mem.SETREGB(regs, reg, bVal);
        break;


      case 0x7E: // REF b
        bVal = Mem.GETREGB(regs, reg);
        Log.fine("REF value", bVal);
        result = ((~bVal) & 2) | (bVal & 1);
        if ((bVal & 4) != 0) result |= 0x80000000;
        if ((bVal & 8) != 0) {
          signedmath = true;
        }
        else {
          signedmath = false;
        }
        break;


      case 0x7F: // RET b,=0
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 4); // pop return address
        pc = Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5));
        result = Mem.SETREGB(regs, reg, (byte)0);
        break;

      case 0x80: // RET w,=0
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 4); // pop return address
        pc = Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5));
        result = Mem.SETREGW(regs, reg, (short)0);
        break;

      case 0x81: // RET d,=0
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 4); // pop return address
        pc = Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5));
        result = Mem.SETREGD(regs, reg, 0);
        break;

      case 0x82: // RET b,=1
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 4); // pop return address
        pc = Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5));
        result = Mem.SETREGB(regs, reg, (byte)1);
        break;

      case 0x83: // RET w,=1
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 4); // pop return address
        pc = Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5));
        result = Mem.SETREGW(regs, reg, (short)1);
        break;

      case 0x84: // RET d,=1
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 4); // pop return address
        pc = Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5));
        result = Mem.SETREGD(regs, reg, 1);
        break;


      case 0x85: // INC b
        if (signedmath) {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) + 1));
        }
        else {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) + 1));
        }
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x86: // INC w
        if (signedmath) {
          result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) + 1));
        }
        else {
          result = Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) + 1));
        }
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x87: // INC d
        if (signedmath) {
          result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) + 1);
        }
        else {
          result = Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) + 1L));
        }
        break;


      case 0x88: // DEC b
        if (signedmath) {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) - 1));
        }
        else {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) - 1));
        }
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x89: // DEC w
        if (signedmath) {
          result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) - 1));
        }
        else {
          result = Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) - 1));
        }
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x8A: // DEC d
        if (signedmath) {
          result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) - 1);
        }
        else {
          result = Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) - 1L));
        }
        break;


      case 0x8B: // NOT b
        result = Mem.SETREGB(regs, reg, (byte)(~Mem.GETREGB(regs, reg)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x8C: // NOT w
        result = Mem.SETREGW(regs, reg, (short)(~Mem.GETREGW(regs, reg)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x8D: // NOT d
        result = Mem.SETREGD(regs, reg, ~Mem.GETREGD(regs, reg));
        break;


      case 0x8E: // NEG b
        result = Mem.SETREGB(regs, reg, (byte)(-Mem.GETREGB(regs, reg)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0x8F: // NEG w
        result = Mem.SETREGW(regs, reg, (short)(-Mem.GETREGW(regs, reg)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0x90: // NEG d
        result = Mem.SETREGD(regs, reg, -Mem.GETREGD(regs, reg));
        break;


      case 0x91: // PUSH b
        Log.fine("push b");
        Mem.SETMEMB(kode, Mem.GETREGD(regs, Mem.REG_5), Mem.GETREGB(regs, reg));
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) + 1);
        break;

      case 0x92: // PUSH w
        Log.fine("push w");
        Mem.SETMEMW(kode, Mem.GETREGD(regs, Mem.REG_5), Mem.GETREGW(regs, reg));
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) + 2);
        break;

      case 0x93: // PUSH d
        Log.fine("push d");
        Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5), Mem.GETREGD(regs, reg));
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) + 4);
        break;


      case 0x94: // POP b
        Log.fine("pop b");
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 1);
        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, Mem.GETREGD(regs, Mem.REG_5)));
        break;

      case 0x95: // POP w
        Log.fine("pop w");
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 2);
        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, Mem.GETREGD(regs, Mem.REG_5)));
        break;

      case 0x96: // POP d
        Log.fine("pop d");
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 4);
        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5)));
        break;


        //************************************************
        // 1-reg imm-direct ops (imm-value always 4 bytes)
        //************************************************


      case 0x98: // LD0 b,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, offset));
        break;

      case 0x99: // LD0 w,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, offset));
        break;

      case 0x9A: // LD0 d,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, offset));
        break;


      case 0x9B: // LD1 b,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_1), offset));
        break;

      case 0x9C: // LD1 w,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_1), offset));
        break;

      case 0x9D: // LD1 d,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_1), offset));
        break;


      case 0x9E: // LD2 b,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_2), offset));
        break;

      case 0x9F: // LD2 w,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_2), offset));
        break;

      case 0xA0: // LD2 d,=dval
        // get offset, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_2), offset));
        break;


      case 0xA1: // LDS b,=dval
        // get offset, moving pc over WORD
        offset = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("offset", offset);

        Mem.SETREGB(regs, reg, Mem.GETMEMB(kode, Mem.GETREGD(regs, Mem.REG_4) + offset));
        break;

      case 0xA2: // LDS w,=dval
        // get offset, moving pc over WORD
        offset = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("offset", offset);

        Mem.SETREGW(regs, reg, Mem.GETMEMW(kode, Mem.GETREGD(regs, Mem.REG_4) + offset));
        break;

      case 0xA3: // LDS d,=dval
        // get offset, moving pc over WORD
        offset = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("offset", offset);

        Mem.SETREGD(regs, reg, Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_4) + offset));
        break;


      case 0xA4: // ST0 b,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETMEMB(kode, offset, Mem.GETREGB(regs, reg));
        break;

      case 0xA5: // ST0 w,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETMEMW(kode, offset, Mem.GETREGW(regs, reg));
        break;

      case 0xA6: // ST0 d,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETMEMD(kode, offset, Mem.GETREGD(regs, reg));
        break;


      case 0xA7: // ST1 b,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_1), offset, Mem.GETREGB(regs, reg));
        break;

      case 0xA8: // ST1 w,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        // set the memory whose handle is in R1
        Mem.SETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_1), offset, Mem.GETREGW(regs, reg));
        break;

      case 0xA9: // ST1 d,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        // get the memory whose handle is in R1
        Mem.SETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_1), offset, Mem.GETREGD(regs, reg));
        break;


      case 0xAA: // ST2 b,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        Mem.SETMEMB(Mem.getMemByRegHandle(regs, Mem.REG_2), offset, Mem.GETREGB(regs, reg));
        break;

      case 0xAB: // ST2 w,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        // set the memory whose handle is in R1
        Mem.SETMEMW(Mem.getMemByRegHandle(regs, Mem.REG_2), offset, Mem.GETREGW(regs, reg));
        break;

      case 0xAC: // ST2 d,=dval
        // get offset of DWORD value, moving pc over DWORD
        offset = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("offset", offset);

        // get the memory whose handle is in R1
        Mem.SETMEMD(Mem.getMemByRegHandle(regs, Mem.REG_2), offset, Mem.GETREGD(regs, reg));
        break;


      case 0xAD: // STS b,=dval
        // get offset, moving pc over WORD
        offset = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("offset", offset);

        Mem.SETMEMB(kode, Mem.GETREGD(regs, Mem.REG_4) + offset, Mem.GETREGB(regs, reg));
        break;

      case 0xAE: // STS w,=dval
        // get offset, moving pc over WORD
        offset = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("offset", offset);

        Mem.SETMEMW(kode, Mem.GETREGD(regs, Mem.REG_4) + offset, Mem.GETREGW(regs, reg));
        break;

      case 0xAF: // STS d,=dval
        // get offset, moving pc over WORD
        offset = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("offset", offset);

        Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_4) + offset, Mem.GETREGD(regs, reg));
        break;


      // going to use common code for IJMP
      case 0xB0: // IJMP b,=dval
      case 0xB1: // IJMP w,=dval
      case 0xB2: // IJMP d,=dval
        if (opcodeinst == 0xB0) {
          dVal = Mem.bUnsigned(Mem.GETREGB(regs, reg));
        }
        else if (opcodeinst == 0xB1) {
          dVal = Mem.wUnsigned(Mem.GETREGW(regs, reg));
        }
        else {
          dVal = Mem.GETREGD(regs, reg);
        }
        Log.fine("ijmp dVal", dVal);

        // next DWORD is offset for the table
        offset = Mem.GETMEMD(kode, pc); 
        Log.fine("offset", offset);
        // don't bother to advance PC, we're jumping

        // get # of entries in table
        count = Mem.GETMEMD(kode, offset);      
        Log.fine("count", count);
        
        // if value too big, then use last enty
        if (dVal > count) {
          dVal = count;
        }

        // also skip "number of entries" DWORD
        ++dVal; 
        
        // skip to the entry (default if past end of table)
        offset += 4*dVal;       
        Log.fine("offset", offset);

        // memory at offset now contains the new pc
        pc = Mem.GETMEMD(kode, offset);
        break;


      // going to use common code for KJMP
      case 0xB3: // KJMP b,=dval
      case 0xB4: // KJMP w,=dval
      case 0xB5: // KJMP d,=dval
        if (opcodeinst == 0xB3) {
          dVal = Mem.bUnsigned(Mem.GETREGB(regs, reg));
        }
        else if (opcodeinst == 0xB4) {
          dVal = Mem.wUnsigned(Mem.GETREGW(regs, reg));
        }
        else {
          dVal = Mem.GETREGD(regs, reg);
        }

        Log.fine("kjmp dVal", dVal);

        // next DWORD is offset for the table
        offset = Mem.GETMEMD(kode, pc); 
        Log.fine("offset", offset);
        // don't bother to advance PC, we're jumping

        // get # of entries in table
        count = Mem.GETMEMD(kode, offset);      

        // jump over count
        offset += 4;

        // while we have an entry left
        while (count-- > 0) {
          // if this is the one
          if (dVal == Mem.GETMEMD(kode, offset) ) {
            // skip value, point at jmp address
            offset += 4;        
            break;
          }

          // skip value & jump address
          offset += 8;  
        }

        // memory at offset now contains the new pc
        pc = Mem.GETMEMD(kode, offset);
        break;


      case 0xB6: // JSR d,=dval
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        offset = Mem.GETREGD(regs, reg);
        Log.fine("offset", offset);

        newpc = dVal + offset;
        Log.fine("newpc", newpc);

        // push return address
        Log.fine("JSR pushing return address to top of stack", pc);
        Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5), pc);

        // move top of stack over this DWORD
        Log.fine("Moving top of stack on a DWORD");
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) + 4);

        // jump to new code
        pc = newpc;
        break;

      case 0xB7: // JSI d,=dval
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        offset = Mem.GETREGD(regs, reg);
        Log.fine("offset", offset);

        newpc = Mem.GETMEMD(kode, offset + dVal);
        Log.fine("newpc", newpc);

        // push return address
        Log.fine("JSI pushing return address to top of stack", pc);
        Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5), pc);

        // move top of stack over this DWORD
        Log.fine("Moving top of stack on a DWORD");
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) + 4);

        // jump to new code
        pc = newpc;
        break;


        //*********************************************************
        // 1-reg imm-value ops
        //*********************************************************

      case 0xBA: // LDI b,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        // set the BYTE register to this value
        Mem.SETREGB(regs, reg, (byte)wVal);

        break;

      case 0xBB: // LDI w,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        // set the WORD register to this value
        Mem.SETREGW(regs, reg, wVal);

        break;

      case 0xBC: // LDI d,=dval
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        // set the DWORD register to this value
        Mem.SETREGD(regs, reg, dVal);

        break;


      case 0xBD: // SHL b,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg)<<(wVal & 0x0007)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0xBE: // SHL w,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg)<<(wVal & 0x000F)));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0xBF: // SHL d,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg)<<(byte)(wVal & 0x001F));
        break;


      case 0xC0: // SHR b,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg)>>>(byte)(wVal & 0x0007)));
        }
        else {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg))>>>(byte)(wVal & 0x0007)));
        }
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0xC1: // SHR w,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg)>>>(byte)(wVal & 0x000F)));
        }
        else {
          result = Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg))>>>(byte)(wVal & 0x000F)));
        }
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0xC2: // SHR d,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg)>>>(byte)(wVal & 0x001F));
        }
        else {
          result = Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg))>>>(byte)(wVal & 0x001F)));
        }
        break;


      case 0xC3: // ROL b,=wval
        // get WORD value from memory, move pc over it
        count = Mem.GETMEMW(kode, pc) & 0x07;
        pc = pc+2;
        Log.fine("count", count);

        bVal = Mem.GETREGB(regs, reg);
        Log.fine("bVal", bVal);

        while (count-- > 0) {
          if ((bVal & 0x80) != 0) {
            bVal = (byte)((bVal << 1) | 1);
          }
          else {
            bVal = (byte)(bVal << 1);
          }
        }
        result = Mem.SETREGB(regs, reg, bVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0xC4: // ROL w,=wval
        // get WORD value from memory, move pc over it
        count = Mem.GETMEMW(kode, pc) & 0x1F;
        pc = pc+2;
        Log.fine("count", count);

        wVal = Mem.GETREGW(regs, reg);
        Log.fine("wVal", wVal);

        while (count-- > 0) {
          if ((wVal & 0x8000) != 0) {
            wVal = (short)((wVal << 1) | 1);
          }
          else {
            wVal = (short)(wVal << 1);
          }
        }
        result = Mem.SETREGW(regs, reg, wVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0xC5: // ROL d,=wval
        // get WORD value from memory, move pc over it
        count = Mem.GETMEMW(kode, pc) & 0x1F;
        pc = pc+2;
        Log.fine("count", count);

        dVal = Mem.GETREGD(regs, reg);
        Log.fine("dVal", dVal);

        while (count-- > 0) {
          if ((dVal & 0x80000000) != 0) {
            dVal = (dVal << 1) | 1;
          }
          else {
            dVal = dVal << 1;
          }
        }
        result = Mem.SETREGD(regs, reg, dVal);
        break;


      case 0xC6: // ROR b,=wval
        // get WORD value from memory, move pc over it
        count = Mem.GETMEMW(kode, pc) & 0x07;
        pc = pc+2;
        Log.fine("count", count);

        bVal = Mem.GETREGB(regs, reg);
        Log.fine("bVal", bVal);

        while (count-- > 0) {
          if ((bVal & 0x01) != 0) {
            bVal = (byte)((bVal >>> 1) | 0x80);
          }
          else {
            bVal = (byte)(bVal >>> 1);
          }
        }
        result = Mem.SETREGB(regs, reg, bVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0xC7: // ROR w,=wval
        // get WORD value from memory, move pc over it
        count = Mem.GETMEMW(kode, pc) & 0x1F;
        pc = pc+2;
        Log.fine("count", count);

        wVal = Mem.GETREGW(regs, reg);
        Log.fine("wVal", wVal);

        while (count-- > 0) {
          if ((wVal & 0x01) != 0) {
            wVal = (short)((wVal >>> 1) | 0x8000);
          }
          else {
            wVal = (short)(wVal >>> 1);
          }
        }
        result = Mem.SETREGW(regs, reg, wVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0xC8: // ROR d,=wval
        // get WORD value from memory, move pc over it
        count = Mem.GETMEMW(kode, pc) & 0x1F;
        pc = pc+2;
        Log.fine("count", count);

        dVal = Mem.GETREGD(regs, reg);
        Log.fine("dVal", dVal);

        while (count-- > 0) {
          if ((dVal & 0x01) != 0) {
            dVal = (dVal >>> 1) | 0x80000000;
          }
          else {
            dVal = dVal >>> 1;
          }
        }
        result = Mem.SETREGD(regs, reg, dVal);
        break;


      case 0xC9: // ADD b,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          // the java default
          Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) + (byte)wVal));
        }
        else {
          Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) + Mem.bUnsigned((byte)wVal)));
        }
        break;

      case 0xCA: // ADD w,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) + wVal));
        }
        else {
          Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) + Mem.wUnsigned(wVal)));
        }
        break;

      case 0xCB: // ADD d,=dval
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        if (signedmath) {
          Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) + dVal);
        }
        else {
          Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) + Mem.dUnsigned(dVal)));
        }
        break;


      case 0xCC: // SUB b,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          // the java default
          result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) - (byte)wVal));
        }
        else {
          result = Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) - Mem.bUnsigned((byte)wVal)));
        }
        break;

      case 0xCD: // SUB w,=wval
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) - wVal));
        }
        else {
          result = Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) - Mem.wUnsigned(wVal)));
        }
        break;

      case 0xCE: // SUB d,=dval
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        if (signedmath) {
          result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) - dVal);
        }
        else {
          result = Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) - Mem.dUnsigned(dVal)));
        }
        break;


      case 0xCF: // MUL b,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          // the java default
          Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) * (byte)wVal));
        }
        else {
          Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) * Mem.bUnsigned((byte)wVal)));
        }
        break;

      case 0xD0: // MUL w,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) * wVal));
        }
        else {
          Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) * Mem.wUnsigned(wVal)));
        }
        break;
        
      case 0xD1: // MUL d,=val
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        if (signedmath) {
          Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) * dVal);
        }
        else {
          Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) * Mem.dUnsigned(dVal)));
        }
        break;

      case 0xD2: // DIV b,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          // the java default
          Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) / (byte)wVal));
        }
        else {
          Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) / Mem.bUnsigned((byte)wVal)));
        }
        break;

      case 0xD3: // DIV w,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) / wVal));
        }
        else {
          Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) / Mem.wUnsigned(wVal)));
        }
        break;

      case 0xD4: // DIV d,=val
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        if (signedmath) {
          Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) / dVal);
        }
        else {
          Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) / Mem.dUnsigned(dVal)));
        }
        break;


      case 0xD5: // MOD b,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          // the java default
          Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) % (byte)wVal));
        }
        else {
          Mem.SETREGB(regs, reg, (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) % Mem.bUnsigned((byte)wVal)));
        }
        break;


      case 0xD6: // MOD w,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) % wVal));
        }
        else {
          Mem.SETREGW(regs, reg, (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) % Mem.wUnsigned(wVal)));
        }
        break;

      case 0xD7: // MOD d,=val
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        if (signedmath) {
          Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) % dVal);
        }
        else {
          Mem.SETREGD(regs, reg, (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) % Mem.dUnsigned(dVal)));
        }
        break;


      case 0xD8: // OR b,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) | (byte)wVal));
        break;

      case 0xD9: // OR w,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) | wVal));
        break;

      case 0xDA: // OR d,=val
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) | dVal);
        break;


      case 0xDB: // AND b,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) & (byte)wVal));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0xDC: // AND w,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) & wVal));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0xDD: // AND d,=val
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) & dVal);
        break;


      case 0xDE: // XOR b,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = Mem.SETREGB(regs, reg, (byte)(Mem.GETREGB(regs, reg) ^ (byte)wVal));
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0xDF: // XOR w,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = Mem.SETREGW(regs, reg, (short)(Mem.GETREGW(regs, reg) ^ wVal));
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0xE0: // XOR d,=val
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        result = Mem.SETREGD(regs, reg, Mem.GETREGD(regs, reg) ^ dVal);
        break;


      case 0xE1: // CMP b,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          result = (byte)(Mem.GETREGB(regs, reg) - (byte)wVal);
        }
        else {
          result = (byte)(Mem.bUnsigned(Mem.GETREGB(regs, reg)) - Mem.bUnsigned((byte)wVal));
        }
        break;

      case 0xE2: // CMP w,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        if (signedmath) {
          result = (short)(Mem.GETREGW(regs, reg) - wVal);
        }
        else {
          result = (short)(Mem.wUnsigned(Mem.GETREGW(regs, reg)) - Mem.wUnsigned(wVal));
        }
        break;

      case 0xE3: // CMP d,=val
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        if (signedmath) {
          result = Mem.GETREGD(regs, reg) - dVal;
        }
        else {
          result = (int)(Mem.dUnsigned(Mem.GETREGD(regs, reg)) - Mem.dUnsigned(dVal));
        }
        break;


      case 0xE4: // TST b,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = (byte)(Mem.GETREGB(regs, reg) & (byte)wVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<24);
        break;

      case 0xE5: // TST w,=val
        // get WORD value from memory, move pc over it
        wVal = Mem.GETMEMW(kode, pc);
        pc = pc+2;
        Log.fine("value", wVal);

        result = (short)(Mem.GETREGW(regs, reg) & wVal);
        // not needed for Java - Java does in above assignment
        // result |= (result<<16);
        break;

      case 0xE6: // TST d,=val
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        result = Mem.GETREGD(regs, reg) & dVal;
        break;


      case 0xE7: // JMP =val
        // get DWORD value from memory (no need to move over it)
        dVal = Mem.GETMEMD(kode, pc);
        // pc = pc+4;
        Log.fine("value", dVal);

        // jump to that location
        pc = dVal;
        break;




        //**************************************************
        // conditional jumps (offset embedded in opcode word
        //**************************************************

      case 0xE8: // JE  +offset (0 to 255 WORDS from CP)
        if (result == 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          pc += (jmpoffset*2);
        }
        break;

      case 0xE9: // JE  -offset (-256 to -1 WORDS from CP)
        if (result == 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
          pc += ((jmpoffset | 0xFFFFFE00)*2);
        }
        break;

      case 0xEA: // JNE +offset
        if (result != 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          pc += (jmpoffset*2);
        }
        break;

      case 0xEB: // JNE -offset
        if (result != 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
          pc += ((jmpoffset | 0xFFFFFE00)*2);
        }
        break;

      case 0xEC: // JL  +offset
        if ((result & 0x80000000) != 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          pc += (jmpoffset*2);
        }
        break;

      case 0xED: // JL  -offset
        if ((result & 0x80000000) != 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
          pc += ((jmpoffset | 0xFFFFFE00)*2);
        }
        break;

      case 0xEE: // JLE +offset
        if (result == 0 || (result & 0x80000000) != 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          pc += (jmpoffset*2);
        }
        break;

      case 0xEF: // JLE -offset
        if (result == 0 || (result & 0x80000000) != 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
          pc += ((jmpoffset | 0xFFFFFE00)*2);
        }
        break;

      case 0xF0: // JG  +offset
        if (result != 0 && (result & 0x80000000) == 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          pc += (jmpoffset*2);
        }
        break;

      case 0xF1: // JG  -offset
        if (result != 0 && (result & 0x80000000) == 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
          pc += ((jmpoffset | 0xFFFFFE00)*2);
        }
        break;

      case 0xF2: // JGE +offset
        if ((result & 0x80000000) == 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          pc += (jmpoffset*2);
        }
        break;

      case 0xF3: // JGE -offset
        if ((result & 0x80000000) == 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
          pc += ((jmpoffset | 0xFFFFFE00)*2);
        }
        break;

      case 0xF4: // JEV +offset
        if ((result & 1) == 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          pc += (jmpoffset*2);
        }
        break;

      case 0xF5: // JEV -offset
        if ((result & 1) == 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
          pc += ((jmpoffset | 0xFFFFFE00)*2);
        }
        break;

      case 0xF6:// JOD +offset
        if ((result & 1) != 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          pc += (jmpoffset*2);
        }
        break;

      case 0xF7: // JOD -offset
        if ((result & 1) != 0) {
          jmpoffset = (opcode & 0x01FF);
          Log.fine("jump offset", jmpoffset);
          Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
          pc += ((jmpoffset | 0xFFFFFE00)*2);
        }
        break;

      case 0xF8: // JMP +offset
        jmpoffset = (opcode & 0x01FF);
        Log.fine("jump offset", jmpoffset);
        pc += (jmpoffset*2);
        break;

      case 0xF9: // JMP -offset
        jmpoffset = (opcode & 0x01FF);
        Log.fine("jump offset", jmpoffset);
        Log.fine("making it negative", (jmpoffset | 0xFFFFFE00));
        pc += ((jmpoffset | 0xFFFFFE00)*2);
        break;



        //************************************************
        // no-reg ops
        //************************************************



      case 0xFA: // SysFunc
        SysFuncs.sysFunc(regs, opcode & 0xff, kode);

        // slug here - so can see graphic operations built up slowly in
        // directdraw
        // System.out.print("Slug in...");
        // for(int slug1=0; slug1<1000000; slug1++) {
        //   for(int slug2=0; slug2<10; slug2++) {
        //   }
        // }
        // System.out.println("Slug out");

        break;


      case 0xFB: // JSR =dval
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        // push return address
        Log.fine("JSR pushing return address to top of stack", pc);
        Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5), pc);

        // move top of stack over this DWORD
        Log.fine("Moving top of stack on a DWORD");
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) + 4);

        // jump to new code
        pc = dVal;
        break;


      case 0xFC: // JSI =dval
        // get DWORD value from memory, move pc over it
        dVal = Mem.GETMEMD(kode, pc);
        pc = pc+4;
        Log.fine("value", dVal);

        // push return address
        Log.fine("JSI pushing return address to top of stack", pc);
        Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5), pc);

        // move top of stack over this DWORD
        Log.fine("Moving top of stack on a DWORD");
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) + 4);

        // jump to new code
        pc = Mem.GETMEMD(kode, dVal);
        break;


      case 0xFD: // RET
        Log.fine("ret: pop return address");
        Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) - 4); 
        pc = Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5));
        break;


      case 0xFE: // misc
        int miscOp = opcode & 0xff;

        // debugging which opcodes are used
        // if (!usedFEOps[miscOp]) {
        //   System.out.println("FE opcode "+Integer.toHexString(miscOp)+" NOT USED BEFORE");
        // usedFEOps[miscOp] = true;
        // }

        switch (miscOp) {

        case 0x00: // BRK
          // for debugging purposes
          Log.info("In break debug");
          KInt.printMainRegisters();
          break;

        case 0x01: // NOP
          break;

        case 0x02: // TERM
          // set flag we need to kill the program
          KInt.rememberKillProgram();
          // NOTE: case TERM falls through into case HALT

        case 0x03: // HALT
          // move pc back to TERM instruction just for safety ???
          pc = pc-2;

          // update PC in regs - not sure this is really needed
          Mem.SETREGD(regs, Mem.REG_3, pc);

          // halt this execution
          haltRequired = true;
          break;

        case 0x04: // AFR
          // pu0 R4,R5
          Mem.SETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5), Mem.GETREGD(regs, Mem.REG_4));
          Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5)+4);

          // ldr R4,R5
          Mem.SETREGD(regs, Mem.REG_4, Mem.GETREGD(regs, Mem.REG_5));

          // add R5,=*PC++
          Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_5) + 
                                       Mem.GETMEMW(kode, pc));
          pc = pc+2;
          break;

        case 0x05: // DFR
          // ldr R5,R4
          // minus the "pop" that's coming up
          Mem.SETREGD(regs, Mem.REG_5, Mem.GETREGD(regs, Mem.REG_4) - 4);

          // po0 R4,R5
          Mem.SETREGD(regs, Mem.REG_4, Mem.GETMEMD(kode, Mem.GETREGD(regs, Mem.REG_5)));
          break;

        case 0x06: // USM
          signedmath = false;
          Log.fine("Using unsigned maths");
          break;

        case 0x07: // SM
          signedmath = true;
          Log.fine("Using signed maths");
          break;

          
        case 0x10: // LDCM b,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))) - 1));
          Mem.SETREGB(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGB(regs, Mem.bUnsigned(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))));
          break;

        case 0x0011: // LDCM w,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))) - 2));
          Mem.SETREGW(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGW(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))))));
          break;

        case 0x0012: // LDCM d,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))) - 4));
          Mem.SETREGD(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGD(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))))));
          break;

        case 0x0013: // LDC b,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGB(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))))));
          break;
                
        case 0x0014: // LDC w,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGW(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGW(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))))));
          break;
                
        case 0x0015: // LDC d,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGD(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGD(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))))));
          break;
                
        case 0x0016: // LDCP b,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGB(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))))));
          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))+1));
          break;

        case 0x0017: // LDCP w,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGW(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGW(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))))));
          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))+2));
          break;

        case 0x0018: // LDCP d,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGD(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF)), Mem.GETREGD(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))))));
          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))+4));
          break;

        case 0x0019: // STCM b,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))-1));
          Mem.SETREGB(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGB(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          break;

        case 0x001A: // STCM w,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))-2));
          Mem.SETREGW(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGW(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          break;
                
        case 0x001B: // STCM d,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))-4));
          Mem.SETREGD(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGD(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          break;

        case 0x001C: // STC b,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGB(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          break;
                
        case 0x001D: // STC w,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGW(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGW(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          break;

        case 0x001E: // STC d,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGD(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGD(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          break;

        case 0x001F: // STCP b,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGB(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGB(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))+1));
          break;
                
        case 0x0020: // STCP w,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGW(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGW(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))+2));
          break;

        case 0x0021: // STCP d,b
          // get WORD value from memory, move pc over it
          reg = Mem.GETMEMW(kode, pc);
          pc = pc+2;
          Log.fine("reg", reg);

          Mem.SETREGD(regs, Mem.bUnsigned((byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF))))), Mem.GETREGD(regs, Mem.bUnsigned((byte)((reg >>> 8) & 0x00FF))));
          Mem.SETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)), (byte)(Mem.GETREGB(regs, Mem.bUnsigned((byte)(reg & 0x00FF)))+4));
          break;

        default:
          Log.errorExit("Unknown misc operation", miscOp);
        }
        break;

      default:
        Log.errorExit("Unknown opcode instruction", opcodeinst);
      }
    }

    // return register 39
    int reg39 = Mem.GETREGW(regs, Mem.REG_39);
    return reg39;
  }




  // Return our CVS id
  public static String getCVSID() {
    return "$Id: ExecKE.java,v 1.6 2002/11/28 19:32:31 Bennett Stephen Exp $";
  }


}


/*
 * $Log: ExecKE.java,v $
 * Revision 1.6  2002/11/28 19:32:31  Bennett Stephen
 * release candidate 1 version
 *
 * Revision 1.5  2002/10/12 14:25:45  Bennett Stephen
 * result variable now static
 *
 * Revision 1.4  2002/10/06 15:29:38  Bennett Stephen
 * CVS method added and some simple changes
 *
 * Revision 1.3  2002/09/08 19:42:22  Bennett Stephen
 * corrected some pc advancements (from 24 to 4)
 *
 * Revision 1.2  2002/08/10 08:23:31  Bennett Stephen
 * memblock 1 and 2 ops now done
 *
 * Revision 1.1  2002/08/07 19:22:22  Bennett Stephen
 * Initial Version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
