// A Java version for the Kaser Interpreter - holds the files found for FindFirst
// For some of my own fun!
// Stephen M. Bennett
// 05Oct02 Started

package stephen.bennett.kint;

import java.io.File;


public class FindFirstFiles {
  // the Files themselves
  private File[] files;

  // index of the current file
  private int currentFileIndex;

  // and remember the file (for effeciency)
  private File currentFile;

  // our handle (so we can ask this object, rather than go back to FileIO)
  private int handle;


  // the constructor
  public FindFirstFiles(File[] fs) {
    files = fs;
    currentFileIndex = 0;
    setCurrentFile();
  }


  // increment the index, returns false if none left
  public boolean increment() {
    if (currentFileIndex == files.length-1) {
      return false;
    }
    currentFileIndex++;
    setCurrentFile();
    return true;
  }


  // set the current file according to the index, does check
  private void setCurrentFile() {
    if (currentFileIndex < 0 || currentFileIndex >= files.length) {
      Log.errorExit("Corrupt index in FindFirstFiles", currentFileIndex);
    }
    
    // okay
    currentFile = files[currentFileIndex];

    // check its defined
    if (currentFile == null) {
      Log.errorExit("null file given to FindFirstFiles, index "+currentFileIndex);
    }
  }


  // our handle will be given later
  public void setHandle(int h) {
    handle = h;
  }
  

  // get our handle
  public int getHandle() {
    return handle;
  }


  // get the attributes of the current file
  // 0x00 for normal file, 0x10 for directory
  public byte getCurrentAttributes() {
    if (currentFile.isFile()) {
      return (byte)0x00;
    }

    if (currentFile.isDirectory()) {
      return (byte)0x10;
    }
    
    // strange
    Log.errorExit("FindFirstFiles: Could not determine current file attribute. Index="+
                  currentFileIndex+" name="+getCurrentName());

    return 0;
  }

  // return the (short) name of the current file
  public String getCurrentName() {
    return currentFile.getName();
  }


  // Return our CVS id
  public static String getCVSID() {
    return "$Id: FindFirstFiles.java,v 1.1 2002/10/06 15:31:09 Bennett Stephen Exp $";
  }



}


/*
 * $Log: FindFirstFiles.java,v $
 * Revision 1.1  2002/10/06 15:31:09  Bennett Stephen
 * intial working version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
