// A Java version for the Kaser Interpreter - The Window Listener
// For some of my own fun!
// Stephen M. Bennett
// 29Jun02 Started

package stephen.bennett.kint;

import java.awt.event.*;


// Map interesting window events to calls on KE Event

class KWindowListener extends WindowAdapter {
  // the window has been closed, calls KE Event shutdown, tidy up and exit
  // this is where TERM eventually ends up, as window is destroyed
  public void windowClosed(WindowEvent e) {
    Log.fine("Window Closed event happened");

    // call the .KE shutdown function
    Log.info("Calling KEEVENT_SHUT_DOWN");
    KInt.keEvent(KInt.KEEVENT_SHUT_DOWN, 0, 0, 0);

    // this was an interesting event (perhaps not, but what the hey)
    UI.setMostRecentEventTime();

    // kill the timer if the Kode failed to
    UI.deleteTimer();

    // okay let's go home
    KInt.killProgram();
  }


  // the user wants to close the window
  public void windowClosing(WindowEvent e) {
    Log.fine("Window Closing event happened");

    // someone has requested the closing of the program.
    // if KEevent(KHDR_QUERY_QUIT) returns 0, don't quit program
    //   else go ahead and quit
    Log.info("Calling KEEVENT_QUERY_QUIT");
    int retVal = KInt.keEvent(KInt.KEEVENT_QUERY_QUIT, 0, 0, 0);
    Log.info("KEVENT_QUERY_QUIT returned", retVal);

    // this was an interesting event
    UI.setMostRecentEventTime();

    if (retVal == 0) {
      // not quiting
      Log.fine("Not quitting");
    }
    else {
      Log.fine("quitting");

      // dispose of the window
      e.getWindow().dispose();
    }
  }



  // the window has been minimized, call KE Event
  public void windowIconified(WindowEvent e) {
    Log.fine("Window Iconified (minimized) event happened");

    // call the .KE function
    Log.info("Calling KEEVENT_WINDOW_HIDDEN");
    KInt.keEvent(KInt.KEEVENT_WINDOW_HIDDEN, 1, 0, 0);

    // this was an interesting event
    UI.setMostRecentEventTime();
  }


  // the window has been maximized, call KE Event
  public void windowDeiconified(WindowEvent e) {
    Log.fine("Window Deiconified (maximized) event happened");

    // call the .KE function
    Log.info("Calling KEEVENT_WINDOW_HIDDEN");
    KInt.keEvent(KInt.KEEVENT_WINDOW_HIDDEN, 0, 0, 0);

    // this was an interesting event
    UI.setMostRecentEventTime();
  }



  // Return our CVS id
  public static String getCVSID() {
    return "$Id: KWindowListener.java,v 1.4 2002/11/28 19:32:30 Bennett Stephen Exp $";
  }


}


/*
 * $Log: KWindowListener.java,v $
 * Revision 1.4  2002/11/28 19:32:30  Bennett Stephen
 * release candidate 1 version
 *
 * Revision 1.3  2002/10/12 14:24:07  Bennett Stephen
 * now call killProgram to exit
 *
 * Revision 1.2  2002/10/06 15:27:02  Bennett Stephen
 * CVS method added
 *
 * Revision 1.1  2002/08/07 19:22:20  Bennett Stephen
 * Initial Version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
