// A Java version for the Kaser Interpreter - the System Functions
// For some of my own fun!
// Stephen M. Bennett
// 29Jun02 Started

package stephen.bennett.kint;

import java.awt.Dimension;
import java.util.Date;
import java.util.Calendar;
import java.text.SimpleDateFormat;


public class SysFuncs {
  // the system function numbers
  private final static int SYS_FILESIZE          = 0x00;
  private final static int SYS_FOPEN             = 0x01;
  private final static int SYS_FEOF              = 0x02;
  private final static int SYS_FREAD             = 0x03;
  private final static int SYS_FWRITE            = 0x04;
  private final static int SYS_FSEEK             = 0x05;
  private final static int SYS_FCLOSE            = 0x06;
  private final static int SYS_FFINDFIRST        = 0x07;
  private final static int SYS_FFINDNEXT         = 0x08;
  private final static int SYS_FFINDCLOSE        = 0x09;
  private final static int SYS_DELETE            = 0x0A;
  private final static int SYS_MAKESHARECOPY     = 0x0B;
  private final static int SYS_PRINTSTART        = 0x0C;
  private final static int SYS_PRINTBEGPAGE      = 0x0D;
  private final static int SYS_PRINTENDPAGE      = 0x0E;
  private final static int SYS_PRINTSTOP         = 0x0F;
  private final static int SYS_ALLOCMEM          = 0x10;
  private final static int SYS_FREEMEM           = 0x11;
  private final static int SYS_TIMEREG           = 0x12;
  private final static int SYS_TIMERSTR          = 0x13;
  private final static int SYS_TIMEMILLI         = 0x14;
  private final static int SYS_STRCAT            = 0x15;
  private final static int SYS_MEMCPY            = 0x16;
  private final static int SYS_MEMMOVE           = 0x17;
  private final static int SYS_STRCPY            = 0x18;
  private final static int SYS_STRCMP            = 0x19;
  private final static int SYS_STRICMP           = 0x1A;
  private final static int SYS_STRLEN            = 0x1B;
  private final static int SYS_MEMSET            = 0x1C;
  private final static int SYS_MEMCMP            = 0x1D;
  private final static int SYS_SETCURSOR         = 0x1E;
  private final static int SYS_SETPGMTITLE       = 0x1F;
  private final static int SYS_CREBM             = 0x20;
  private final static int SYS_DELBM             = 0x21;
  private final static int SYS_ALPHABLEND        = 0x22;  
  private final static int SYS_SIZEBM            = 0x23;
  private final static int SYS_DRAWBM            = 0x24;
  private final static int SYS_STRETCHBM         = 0x25;
  private final static int SYS_DRAWMBM           = 0x26;
  private final static int SYS_LOADBM            = 0x27;
  private final static int SYS_LOADMBM           = 0x28;
  private final static int SYS_STRETCHMBM        = 0x29;  
  private final static int SYS_GETWINDOWPOS      = 0x2A;
  private final static int SYS_SETWINDOWPOS      = 0x2B;
  private final static int SYS_GETVIEWSIZE       = 0x2C;
  private final static int SYS_ORIENTBM          = 0x2D;  
  private final static int SYS_GETSCREENSIZE     = 0x2E;
  private final static int SYS_FLUSHSCREEN       = 0x2F;
  private final static int SYS_RECTEDGED         = 0x30;
  private final static int SYS_RECTSOLID         = 0x31;
  private final static int SYS_RECTOUTLINE       = 0x32;
  private final static int SYS_LINES             = 0x33;
  private final static int SYS_SETPIXEL          = 0x34;
  private final static int SYS_GETPIXEL          = 0x35;
  private final static int SYS_COLORIZE          = 0x36;  
  private final static int SYS_RECT              = 0x37;
  private final static int SYS_OVAL              = 0x38;  
  private final static int SYS_GETTEXTSIZE       = 0x39;
  private final static int SYS_GETFONT           = 0x3A;
  private final static int SYS_RELFONT           = 0x3B;
  private final static int SYS_BEGTEXT           = 0x3C;
  private final static int SYS_TEXTSTYLE         = 0x3D;
  private final static int SYS_TEXT              = 0x3E;
  private final static int SYS_ENDTEXT           = 0x3F;
  private final static int SYS_PLAYSOUND         = 0x40;
  private final static int SYS_PLAYMUSIC         = 0x41;
  private final static int SYS_STOPMUSIC         = 0x42;
  private final static int SYS_ISMUSIC           = 0x43;
  private final static int SYS_MEMSET2           = 0x44;
  private final static int SYS_MEMSET4           = 0x45;
  private final static int SYS_PRINTFONT         = 0x47;
  private final static int SYS_STARTTIMER        = 0x48;
  private final static int SYS_STOPTIMER         = 0x49;
  private final static int SYS_GETEOL            = 0x4A;
  private final static int SYS_PRINTGETTEXTSIZE  = 0x4B;
  private final static int SYS_PRINTTEXT         = 0x4C;
  private final static int SYS_PRINTBM           = 0x4D;
  private final static int SYS_PRINTLINES        = 0x4E;
  private final static int SYS_SYSTEMMSG         = 0x4F;
  private final static int SYS_LOADKE            = 0x50;
  private final static int SYS_UNLOADKE          = 0x51;
  private final static int SYS_MSGKE             = 0x52;
  private final static int SYS_INVDRAWBM         = 0x55;
  private final static int SYS_INVDRAWMBM        = 0x56;
  private final static int SYS_LINESTYLE         = 0x57;
  private final static int SYS_MAKEFOLDER        = 0x58;
  private final static int SYS_DELFOLDER         = 0x59;
  private final static int SYS_CLIP              = 0x5A;
  private final static int SYS_GETCLIP           = 0x5B;
  private final static int SYS_MASKBM            = 0x5C;
  private final static int SYS_RGB2HSL           = 0x5D;
  private final static int SYS_HSL2RGB           = 0x5E;
  private final static int SYS_SAVEBM            = 0x5F;
  private final static int SYS_PRINTRECT         = 0x60;
  

  // the first time, so can provide a progessive millisecond count
  private static long startTime;

  // the new line separator
  private static String newlineSep = System.getProperty("line.separator");

  // the date format for SysTimeStr
  private static SimpleDateFormat sdf = new SimpleDateFormat("EEE MMM dd HH:mm:ss yyyy");

  
  // set the start time to now
  public static void setStartTime() {
    startTime = System.currentTimeMillis();
  }


  // call the named system function, registers and kode supplied supplied
  public static void sysFunc(byte[] regs, int sysfunc, byte[] kode) {
    Log.info("calling system func", sysfunc);

    switch(sysfunc) {


    //****************************************
    // Debug Functions
    //****************************************

    case SYS_SYSTEMMSG: {
      Log.info("SYS_SYSTEMMSG");

      // get the string at that r15 location
      String s = Mem.getString(kode, Mem.GETREGD(regs, Mem.REG_15));

      // show this string as a modal message dialog
      UI.showMessageDialog(s);

      return;
    }


    //****************************************
    // Timer Functions
    //****************************************

    case SYS_STARTTIMER: {
      // inputs:
      // R15 = timer period in milliseconds

      Log.info("SYS_STARTTIMER");

      // create periodic timer
      UI.createTimer(Mem.GETREGD(regs, Mem.REG_15));

      return;
    }


    case SYS_STOPTIMER: {
      Log.info("SYS_STOPTIMER");

      // delete the periodic timer
      UI.deleteTimer();

      return;
    }

      
    case SYS_TIMEMILLI: {
      // Outputs:
      //   R15 = "progressive" time count in milliseconds.
      // By "progressive", I mean a count which doesn't necessarily bear
      // any resemblance to "real" time, but rather is a counter that is
      // a reasonably accurate "millisecond counter" that may eventually
      // roll over to 0, but which should not do so more often than once a day,
      // and who's count value continues upwards (until such time as a roll-over
      // might occur).  This is accomplished on the MS-Windows platform using
      // the GetTickCount() function.  Other platforms may choose to implement
      // this in whatever way seems easiest, while meeting the above criteria.
      // This counter value is only used for timing the speed of animations
      // and brief delays in actions.  So, reasonable accuracy down to 5 or 10
      // milliseconds would be very nice, and the roll-over (if and when it
      // occurs) should only cause (at most) a slight burp in any animation
      // or screen action that is going on at the moment (but from which it
      // should recover almost immediately).
      // NOTE: It is IMPORTANT that his count be relative accurate, as it is
      //       used for generating the elapsed time count for solving puzzles!

      Log.info("SYS_TIMEMILLI");

      // number of millisecs since we started, this will roll over in
      // 24 days.
      Mem.SETREGD(regs, Mem.REG_15, (int)((System.currentTimeMillis()-startTime)&0xFFFFFFFF));

      return;
    }


    //****************************************
    // Time Functions
    //****************************************

    case SYS_TIMEREG: {
      // Outputs:
      //        R39 = year (ie, 2001)
      //        R38 = month (0-11, Jan=0, Feb=1, Mar=2, etc)
      //        R37 = day (1-31)
      //        R36 = hour (0-23)
      //        R35 = minute (0-59)
      //        R34 = second (0-59)

      Log.info("SYS_TIMERREG");

      // create a calendar for now
      Calendar now = Calendar.getInstance();

      // set the regs accordingly
      Mem.SETREGW(regs, Mem.REG_39, (short)now.get(Calendar.YEAR));
      Mem.SETREGW(regs, Mem.REG_38, (short)now.get(Calendar.MONTH));
      Mem.SETREGW(regs, Mem.REG_37, (short)now.get(Calendar.DATE));
      Mem.SETREGW(regs, Mem.REG_36, (short)now.get(Calendar.HOUR_OF_DAY));
      Mem.SETREGW(regs, Mem.REG_35, (short)now.get(Calendar.MINUTE));
      Mem.SETREGW(regs, Mem.REG_34, (short)now.get(Calendar.SECOND));

      return;
    }


    case SYS_TIMERSTR: {
      // Inputs:
      //      R15 = offset in memblock[R95] of buffer for TIME string
      //      R95 = 0 for R0, 1 for R1, 2 for R2
      // Outputs:
      //      buffer [R95]:R15 filled with string representing current time

      Log.info("SYS_TIMERSTR");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // get now as a string
      String now = sdf.format(new Date());
      Log.fine("timerstr", now);

      // set the string at that location, including newline separator
      Mem.setString(theMem, offset, now+newlineSep, true);

      return;
    }


    //****************************************
    // Window Functions
    //****************************************

    case SYS_SETPGMTITLE: {
      // Inputs:
      //        R15 = offset in memblock[R95] of buffer containing program title
      //        R95 = 0 for R0, 1 for R1, 2 for R2

      Log.info("SYS_SETPGMTITLE");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // set window text to the string at that memory lo<ion
      UI.setWindowText(Mem.getString(theMem, offset));

      return;
    }


    case SYS_GETWINDOWPOS: {
      // outputs:
      //        R15 = X
      //        R14 = Y
      //        R13 = W
      //        R12 = H
      //        R11 = system dependant  (for me extended state)

      Log.info("SYS_GETWINDOWPOS");

      Mem.SETREGD(regs, Mem.REG_15, UI.getX());
      Mem.SETREGD(regs, Mem.REG_14, UI.getY());
      Mem.SETREGD(regs, Mem.REG_13, UI.getWidth());
      Mem.SETREGD(regs, Mem.REG_12, UI.getHeight());
      int state = UI.getExtendedState();
      //System.out.println("GETWINDOWPOS state="+state);
      Log.fine("state", state);
      Mem.SETREGD(regs, Mem.REG_11, state);

      return;
    }


    case SYS_SETWINDOWPOS: {
      // inputs:
      //        R15 = X
      //        R14 = Y
      //        R13 = W
      //        R12 = H
      //        R11 = system dependant (for me extended state)

      Log.info("SYS_SETWINDOWPOS");

      UI.setLocation(Mem.GETREGD(regs, Mem.REG_15),
                     Mem.GETREGD(regs, Mem.REG_14));
      UI.setSize(Mem.GETREGD(regs, Mem.REG_13),
                 Mem.GETREGD(regs, Mem.REG_12));
      int state = Mem.GETREGD(regs, Mem.REG_11);
      //System.out.println("SETWINDOWPOS state="+state);
      Log.fine("ignoring state", state);
      // UI.setExtendedState(state);

      return;
    }


    case SYS_GETVIEWSIZE: {
      // outputs:
      //        R39 = window width
      //        R38 = window height

      Log.info("SYS_GETVIEWSIZE");

      Mem.SETREGW(regs, Mem.REG_39, (short)UI.getViewWidth());
      Mem.SETREGW(regs, Mem.REG_38, (short)UI.getViewHeight());

      return;
    }


    case SYS_GETSCREENSIZE: {
      // outputs:
      //        R39 = screen width
      //        R38 = screen height
      //        R37 = size index (0=640, 1=800, 2=1024, 3=1152, 4=1280, 5=1600)

      Log.info("SYS_GETSCREENSIZE");
      Mem.SETREGW(regs, Mem.REG_39, (short)UI.getScreenWidth());
      Mem.SETREGW(regs, Mem.REG_38, (short)UI.getScreenHeight());
      Mem.SETREGW(regs, Mem.REG_37, (short)UI.getScreenWidthIndex());

      return;
    }


    case SYS_FLUSHSCREEN: {
      // force display update/refresh RIGHT NOW (rather than waiting a later 
      // "update message" or whatever)

      Log.info("SYS_FLUSHSCREEN");
      UI.paintInvalidArea();

      // Log.info("no repaint needed for java KINT");
      // UI.repaint();

      return;
    }


    //****************************************
    // Bitmap Functions
    //****************************************


    case SYS_SETCURSOR: {
      // Inputs:
      //        R39 = 0 for normal, 1 for wait, 2 for hotlink
      
      Log.info("SYS_SETCURSOR");

      UI.setCursor(Mem.GETREGW(regs, Mem.REG_39));
      UI.showCursor();

      return;
    }

    case SYS_CREBM: {
      // inputs:
      //      R39 = width
      //      R38 = height
      // outputs:
      //      R15 = handle (PDIBS in this implementation) of created bitmap, 
      //            NULL if creation error.

      Log.info("SYS_CREBM");

      // find the w and h
      int w = Mem.GETREGW(regs, Mem.REG_39);
      int h = Mem.GETREGW(regs, Mem.REG_38);

      // create the bitmap returning handle in r15
      Mem.SETREGD(regs, Mem.REG_15, UI.createBMB(w, h, false));

      return;
    }


    case SYS_DELBM: {
      // inputs:
      //        R15 = handle of bitmap to delete

      Log.info("SYS_DELBM");

      int handle = Mem.GETREGD(regs, Mem.REG_15);
      UI.deleteBitmap(handle);

      return;
    }


    case SYS_SIZEBM: {
      // Inputs:
      //      R15 = handle of bitmap
      // Outputs:
      //      R39 = width of bitmap
      //      R38 = height of bitmap
      //	    R37 = orientation of bitmap (BMO_ value) (KINT 2.0)
      
      Log.info("SYS_SIZEBM");

      int handle = Mem.GETREGD(regs, Mem.REG_15);
      
      // get the BMB
      BMB bmb = UI.getBMB(handle);
      Mem.SETREGW(regs, Mem.REG_39, (short)bmb.getLogW());
      Mem.SETREGW(regs, Mem.REG_38, (short)bmb.getLogH());
      Mem.SETREGW(regs, Mem.REG_37, (short)bmb.getOrient());
      return;
    }

    
    case SYS_MASKBM: {
      // Create new masked for BitMap KINT 2.0
      // Inputs:
      //   R15 = bitmap handle
      //   R14 = transparent color 1
      //   R13 = transparent color 2

      /* 
       * CreateMaskBMB((PBMB)GETREGD(REG_15), GETREGD(REG_14), GETREGD(REG_13));
       */
      
      System.out.println("SYS_MASKBM TBD");
      return;
    }
      

    case SYS_DRAWBM: {
      // Inputs:
      //        R15 = handle of destination bitmap
      //        R14 = handle of source bitmap
      //        R39 = x-coord of destination
      //        R38 = y-coord of destination
      //        R37 = x-coord of source
      //        R36 = y-coord of source
      //        R35 = width of source
      //        R34 = height of source

      Log.info("SYS_DRAWBM");

      int handleDst = Mem.GETREGD(regs, Mem.REG_15);
      int handleSrc = Mem.GETREGD(regs, Mem.REG_14);
      int xDst = Mem.GETREGW(regs, Mem.REG_39);
      int yDst = Mem.GETREGW(regs, Mem.REG_38);
      int xSrc = Mem.GETREGW(regs, Mem.REG_37);
      int ySrc = Mem.GETREGW(regs, Mem.REG_36);
      int wSrc = Mem.GETREGW(regs, Mem.REG_35);
      int hSrc = Mem.GETREGW(regs, Mem.REG_34);
      UI.drawBitmap(handleDst, handleSrc, xDst, yDst, xSrc, ySrc, wSrc, hSrc);

      return;
    }


    case SYS_STRETCHBM: {
      // Inputs:
      //        R15 = handle of destination bitmap
      //        R14 = handle of source bitmap
      //        R39 = x-coord of destination
      //        R38 = y-coord of destination
      //        R37 = x-coord of source
      //        R36 = y-coord of source
      //        R35 = width of source
      //        R34 = height of source
      //        R33 = width of destination
      //        R32 = height of destination

      Log.info("SYS_STRETCHBM");

      int handleDst = Mem.GETREGD(regs, Mem.REG_15);
      int handleSrc = Mem.GETREGD(regs, Mem.REG_14);
      int xDst = Mem.GETREGW(regs, Mem.REG_39);
      int yDst = Mem.GETREGW(regs, Mem.REG_38);
      int xSrc = Mem.GETREGW(regs, Mem.REG_37);
      int ySrc = Mem.GETREGW(regs, Mem.REG_36);
      int wSrc = Mem.GETREGW(regs, Mem.REG_35);
      int hSrc = Mem.GETREGW(regs, Mem.REG_34);
      int wDst = Mem.GETREGW(regs, Mem.REG_33);
      int hDst = Mem.GETREGW(regs, Mem.REG_32);
      UI.stretchBitmap(handleDst, handleSrc, xDst, yDst, wDst, hDst,
                       xSrc, ySrc, wSrc, hSrc);

      return;
    }


    case SYS_DRAWMBM: {
      // Inputs:
      //        R15 = handle of destination bitmap
      //        R14 = handle of source bitmap
      //        R39 = x-coord of destination
      //        R38 = y-coord of destination
      //        R37 = x-coord of source
      //        R36 = y-coord of source
      //        R35 = width of source
      //        R34 = height of source

      Log.info("SYS_DRAWMBM");

      int handleDst = Mem.GETREGD(regs, Mem.REG_15);
      int handleSrc = Mem.GETREGD(regs, Mem.REG_14);
      int xDst = Mem.GETREGW(regs, Mem.REG_39);
      int yDst = Mem.GETREGW(regs, Mem.REG_38);
      int xSrc = Mem.GETREGW(regs, Mem.REG_37);
      int ySrc = Mem.GETREGW(regs, Mem.REG_36);
      int wSrc = Mem.GETREGW(regs, Mem.REG_35);
      int hSrc = Mem.GETREGW(regs, Mem.REG_34);
      UI.drawMBitmap(handleDst, handleSrc, xDst, yDst, xSrc, ySrc, wSrc, hSrc);

      return;
    }


    case SYS_LOADBM: {
      // Inputs:
      //        R15 = offset into memory block[R95] of filename
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      //        R94 =
      //                0 file is in GAME folder
      //                1 file is in UI folder
      //                2 file is in IMAGES folder
      //                3 file is in BACKGNDS folder
      // Outputs:
      //        R15 = handle (PDIBS) of loaded bitmap, or NULL if failed

      Log.info("SYS_LOADBM");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // find the file name
      String filename = FileIO.buildPath(Mem.getString(theMem, offset),
                                         Mem.GETREGB(regs, Mem.REG_94), true);

      // load the MBM returning handle in r15
      Mem.SETREGD(regs, Mem.REG_15, UI.loadBitmap(filename));

      return;
    }


    case SYS_LOADMBM: {
      // Inputs:
      //        R15 = offset into memory block[R95] of filename
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      //        R39 = x-coord of transparent pixel
      //        R38 = y-coord of transparent pixel
      //        R94 =
      //                0 file is in GAME folder
      //                1 file is in UI folder
      //                2 file is in IMAGES folder
      //                3 file is in BACKGNDS folder
      // Outputs:
      //        R15 = handle (PDIBS) of loaded bitmap, or NULL if failed

      Log.info("SYS_LOADMBM");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // find x and y co-ords
      int x = Mem.GETREGW(regs, Mem.REG_39);
      int y = Mem.GETREGW(regs, Mem.REG_38);

      // find the file name
      String filename = FileIO.buildPath(Mem.getString(theMem, offset),
                                         Mem.GETREGB(regs, Mem.REG_94), true);

      // load the MBM returning handle in r15
      Mem.SETREGD(regs, Mem.REG_15, UI.loadMBitmap(filename, x, y));

      return;
    }

    
    case SYS_CLIP: {
      // KINT 2.0
      // inputs:
      //   R15 = bitmap handle (NULL==screen) to clip
      //   R39 = left-coordinate of clip region (nothing is drawn < R39)
      //   R38 = top-coordinate of clip region  (nothing is drawn < R38)
      //   R37 = width of clip region (nothing is drawn >= R39+R37)
      //   R36 = height of clip region (nothing is drawn >= R38+R36)
      // NOTE: if R39==-1, then UNclip (set clip to bitmap's size)
      
      /*
      pKB = (PBMB)GETREGD(REG_15);
      if( pKB==0 ) pKB = kBuf;
      if( -1==(pKB->clipX1 = (int)(short)GETREGW(REG_39)) ) {
        pKB->clipX1 = pKB->clipY1 = 0;
        pKB->clipX2 = pKB->logW;
        pKB->clipY2 = pKB->logH;
      } else {
        pKB->clipX1 = max(pKB->clipX1, 0);
        pKB->clipY1 = max((int)(short)GETREGW(REG_38), 0);
        pKB->clipX2 = min(pKB->clipX1 + (int)(short)GETREGW(REG_37), pKB->logW);
        pKB->clipY2 = min(pKB->clipY1 + (int)(short)GETREGW(REG_36), pKB->logH);
      }
      */
     
        System.out.println("SYS_CLIP TBD");
        return;
    }
      
    case SYS_COLORIZE: {
      // KINT 2.0
      // inputs:
      //   R15 = bitmap handle (NULL==screen)
      //   R39 = x
      //   R38 = y
      //   R37 = w
      //   R36 = h
      //   R35 = hue (-1==no change, else 0-255)
      //   R34 = saturation (-1==no change, 0=convert to grayscale, else 1-255.  When sat==0, hue is ignored.)
      //   R33 = brightness (-255 to 255, 0==no change, else newbrightness = oldbrightness + R33 (result limited to 0-255)
      //   R32 = contrast (-255 to 255, 0==no change, else 
    
      /*
      if( GETREGW(REG_34)==0 ) GrayBMB((PBMB)GETREGD(REG_15), (int)(short)GETREGW(REG_39), (int)(short)GETREGW(REG_38),
          (int)(short)GETREGW(REG_37), (int)(short)GETREGW(REG_36), (int)(short)GETREGW(REG_33), (int)(short)GETREGW(REG_32));
      else ColorizeBMB((PBMB)GETREGD(REG_15), (int)(short)GETREGW(REG_39), (int)(short)GETREGW(REG_38),
          (int)(short)GETREGW(REG_37), (int)(short)GETREGW(REG_36), (int)(short)GETREGW(REG_35), (int)(short)GETREGW(REG_34), (int)(short)GETREGW(REG_33), (int)(short)GETREGW(REG_32));
          */
      System.out.println("SYS_COLORIZE TBD");
      return;
    }
      
    case SYS_ORIENTBM: {
      // inputs:
      //   R15 = Bitmap handle
      //   R39 = BMO_ value

      /*
      pKB = (PBMB)GETREGD(REG_15);
      if( pKB==0 ) pKB = kBuf;
      w = GETREGW(REG_39);
      if( (w ^ pKB->orient) & BMO_90 ) {	// if rotating multiple of 90 degrees
        h = pKB->logH;					//	swap logW and logH
        pKB->logH = pKB->logW;
        pKB->logW = h;
      }
      pKB->orient = w;
      pKB->clipX1 = pKB->clipY1 = 0;
      if( w & BMO_90 ) {	// if rotated multiple of 90 degrees
        pKB->clipX2 = pKB->physH;
        pKB->clipY2 = pKB->physW;
      } else {
        pKB->clipX2 = pKB->physW;
        pKB->clipY2 = pKB->physH;
      }
      */
      
      Log.errorExit("I hoped SYS_ORIENTBM was never called for Java");
      return;
    }

    
    //****************************************
    // Graphic Functions
    //****************************************


    case SYS_RECTEDGED: {
      // inputs:
      //        R15 = PDIBS handle (NULL==screen)
      //        R14 = RGB color fill
      //        R13 = RGB color frame
      //        R39 = left
      //        R38 = top
      //        R37 = right
      //        R36 = bottom

      Log.info("SYS_RECTEDGED");

      int handle = Mem.GETREGD(regs, Mem.REG_15);
      int fillColour = Mem.GETREGD(regs, Mem.REG_14);
      int frameColour = Mem.GETREGD(regs, Mem.REG_13);
      short left = Mem.GETREGW(regs, Mem.REG_39);
      short top = Mem.GETREGW(regs, Mem.REG_38);
      short right = Mem.GETREGW(regs, Mem.REG_37);
      short bottom = Mem.GETREGW(regs, Mem.REG_36);
      UI.rectEdged(handle, fillColour, frameColour, left, top, right, bottom);

      return;
    }


    case SYS_RECTSOLID: {
      // inputs:
      //        R15 = PDIBS handle (NULL==screen)
      //        R14 = RGB color fill
      //        R39 = left
      //        R38 = top
      //        R37 = right
      //        R36 = bottom

      Log.info("SYS_RECTSOLID");

      int handle = Mem.GETREGD(regs, Mem.REG_15);
      int colour = Mem.GETREGD(regs, Mem.REG_14);
      short left = Mem.GETREGW(regs, Mem.REG_39);
      short top = Mem.GETREGW(regs, Mem.REG_38);
      short right = Mem.GETREGW(regs, Mem.REG_37);
      short bottom = Mem.GETREGW(regs, Mem.REG_36);
      UI.rectSolid(handle, colour, left, top, right, bottom);

      return;
    }


    case SYS_RECTOUTLINE: {
      // inputs:
      //        R15 = PDIBS handle (NULL==screen)
      //        R13 = RGB color frame
      //        R39 = left
      //        R38 = top
      //        R37 = right
      //        R36 = bottom

      Log.info("SYS_RECTOUTLINE");

      int handle = Mem.GETREGD(regs, Mem.REG_15);
      int colour = Mem.GETREGD(regs, Mem.REG_13);
      short left = Mem.GETREGW(regs, Mem.REG_39);
      short top = Mem.GETREGW(regs, Mem.REG_38);
      short right = Mem.GETREGW(regs, Mem.REG_37);
      short bottom = Mem.GETREGW(regs, Mem.REG_36);
      UI.rectOutline(handle, colour, left, top, right, bottom);

      return;
    }

   case SYS_RECT: { 
     // KINT 2.0 and KINT 2.2
     // inputs:
     //   R15 = bitmap handle (NULL==screen)
     //   R14 = RGB color fill (-1 if none) (on 2.2, if -2, then R90-R94 are valid and do Gradient fill)
     //   R13 = RGB color frame (-1 if none) (on 2.2, if -2, then R90-R94 are valid and do Gradient fill)
     //   R39 = left
     //   R38 = top
     //   R37 = w
     //   R36 = h
     //   R90 = hue (if R13 or R14 == -2)
     //   R91 = satmin (if R13 or R14 == -2)
     //   R92 = satmax (if R13 or R14 == -2)
     //   R93 = lightmin (if R13 or R14 == -2)
     //   R94 = lightmax (if R13 or R14 == -2)
     // WARN: PTR4
    
     /*
      x = (int)(short)GETREGW(REG_39);
      y = (int)(short)GETREGW(REG_38);
      dw= GETREGD(REG_13);
      dh= GETREGD(REG_14);
      if( dw==-2 || dh==-2 ) RectGradientBMB((PBMB)GETREGD(REG_15), x, y, (int)(short)GETREGW(REG_37), (int)(short)GETREGW(REG_36), GETREGB(REG_90), GETREGB(REG_91), GETREGB(REG_92), GETREGB(REG_93), GETREGB(REG_94));
      else RectBMB((PBMB)GETREGD(REG_15), x, y, x+(int)(short)GETREGW(REG_37), y+(int)(short)GETREGW(REG_36), dh, dw);
      */
     
     System.out.println("SYS_RECT TBD");
      return;
   }
      
    
    case SYS_LINES: {
      // inputs:
      //        R15 = PDIBS handle (NULL==screen)
      //        R14 = offset in memblock[R95] of (x,y) WORD coordinate array
      //        R13 = RGB color of lines
      //        R39 = # of points in the [R14] array
      //        R95 = 0 for R0, 1 for R1, 2 for R2 memory block for R14 offset

      Log.info("SYS_LINES");

      int handle = Mem.GETREGD(regs, Mem.REG_15);
      int colour = Mem.GETREGD(regs, Mem.REG_13);
      int nPoints = Mem.GETREGW(regs, Mem.REG_39);

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // create points array - for Java two arrays, one for X and one for Y
      Log.fine("creating points arrays, size", nPoints);
      int[] xs = new int[nPoints];
      int[] ys = new int[nPoints];

      // use the same algorithm for determining the invalidate rectangle
      // as KINT does
      int left = 32000;
      int top = 32000;
      int right = 0;
      int bottom = 0;

      // populate - entries are WORDs
      for(int i=0; i<nPoints; i++) {
        xs[i] = Mem.GETMEMW(theMem, offset);
        offset = offset+2;
        ys[i] = Mem.GETMEMW(theMem, offset);
        offset = offset+2;

        if (xs[i] < left) left = xs[i];
        if (xs[i] > right) right = xs[i];
        if (ys[i] < top) top = ys[i];
        if (ys[i] > bottom) bottom = ys[i];
      }

      // create the invalidate rectangle to pass on
      java.awt.Rectangle r = new java.awt.Rectangle(left, top, right-left, bottom-top);

      UI.drawPolyline(handle, colour, xs, ys, nPoints, r);

      return;
    }

    
    case SYS_LINESTYLE: {
      // KINT 2.2
      // inputs:
      //   R12 = new linestyle (DWORD, bits set where points plotted, gets rotated as used by LineBMB and RectBMB)
      // outputs
      //   R12 = old linestyle
      
      /*
      dw = GETREGD(REG_12);
      SETREGD(REG_12, LineStyle);
      LineStyle = dw;
      */
      System.out.println("SYS_LINESTYLE TDB");
      return;
    }

    
    case SYS_SETPIXEL: {
      // inputs:
      //        R15 = PDIBS handle (NULL==screen)
      //        R14 = RGB color of point
      //        R39 = x-coordinate of point
      //        R38 = y-coordinate of point

      Log.info("SYS_SETPIXEL");

      // get the handle
      int handle = Mem.GETREGD(regs, Mem.REG_15);

      // find rgb, x and y co-ords
      int rgb = Mem.GETREGD(regs, Mem.REG_14);
      int x = Mem.GETREGW(regs, Mem.REG_39);
      int y = Mem.GETREGW(regs, Mem.REG_38);

      // set the pixel
      UI.setPixel(handle, x, y, rgb);

      return;
    }


    case SYS_GETPIXEL: {
      // inputs:
      //        R15 = PDIBS handle (NULL==screen)
      //        R39 = x-coordinate of point
      //        R38 = y-coordinate of point
      // outputs:
      //        R14 = RGB color of point

      Log.info("SYS_GETPIXEL");

      // get the handle
      int handle = Mem.GETREGD(regs, Mem.REG_15);

      // find x and y co-ords
      int x = Mem.GETREGW(regs, Mem.REG_39);
      int y = Mem.GETREGW(regs, Mem.REG_38);

      // get and return the pixel
      Mem.SETREGD(regs, Mem.REG_14, UI.getPixel(handle, x, y));

      return;
    }


    //****************************************
    // Font Functions
    //****************************************

    case SYS_GETFONT: {
      // inputs:
      //        R14 = string buffer address in memblk[R95]
      //        R95 = memblk for string buffer
      //        R39 = pixel width of rectangle to fit text into
      //        R38 = pixel height of rectangle to fit text into
      //        R37 =
      //                BIT 0 =0 for variable pitch, =1 for fixed pitch
      //                BIT 1 =0 for normal, =1 for BOLD
      //                BIT 2 =0 for normal, =1 for ITALIC
      // outputs:
      //        R39 = handle of font
      //        R38 = height of font
      //        R37 = ascending height

      Log.info("SYS_GETFONT");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // get the string at that memory location
      String s = Mem.getString(theMem, offset);
      
      // get the font index that fits this string
      short fontIndex = UI.fitFont(s, Mem.GETREGW(regs, Mem.REG_39), 
                                      Mem.GETREGW(regs, Mem.REG_38), 
                                      Mem.GETREGW(regs, Mem.REG_37));

      // save the information back to registers
      Mem.SETREGW(regs, Mem.REG_39, fontIndex);
      Mem.SETREGW(regs, Mem.REG_38, (short)UI.getFontHeight(fontIndex));
      Mem.SETREGW(regs, Mem.REG_37, (short)UI.getFontAscent(fontIndex));

      return;
    }


    case SYS_RELFONT: {
      // inputs:
      //        R39 = handle of font to release

      Log.info("SYS_RELFONT");
      
      // get and release font by handle number
      UI.releaseFont(Mem.GETREGW(regs, Mem.REG_39));

      return;
    }


    //****************************************
    // Text Functions
    //****************************************

    case SYS_BEGTEXT: {
      // save whatever states you need to save regarding text output for "ddc"
      // so that they can be restored at SYS_ENDTEXT
      // inputs:

      Log.info("SYS_BEGTEXT");

      UI.saveTextState();

      return;
    }


    case SYS_ENDTEXT: {
      // restore whatever states need to be restored that were saved at SYS_BEGTEXT

      Log.info("SYS_ENDTEXT");

      UI.restoreTextState();

      return;
    }


    case SYS_TEXTSTYLE: {
      //TextStyle(backcolor=-1 if transparent, textcolor, font, align)
      // inputs:
      //        R14 = backcolor (-1 if transparent, otherwise opaque)
      //        R13 = textcolor
      //        R39 = font handle

      Log.info("SYS_TEXTSTYLE");

      int textColor = Mem.GETREGD(regs, Mem.REG_13);
      int backColor = Mem.GETREGD(regs, Mem.REG_14);
      short fontHandle = Mem.GETREGW(regs, Mem.REG_39);

      UI.setTextStyle(textColor, backColor, fontHandle);

      return;
    }


    case SYS_TEXT: {
      // inputs:
      //        R15 = PDIBS handle (NULL==screen)
      //        R14 = string buffer address in memblk[R95]
      //        R95 = memblk for string buffer
      //        R39 = x-coordinate of point
      //        R38 = y-coordinate of point

      Log.info("SYS_TEXT");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // get the string at that memory location
      String s = Mem.getString(theMem, offset);

      int pBM = Mem.GETREGD(regs, Mem.REG_15);
      short x = Mem.GETREGW(regs, Mem.REG_39);
      short y = Mem.GETREGW(regs, Mem.REG_38);
      
      UI.drawString(pBM, s, x, y);

      return;
    }


    case SYS_GETTEXTSIZE: {
      // assumes a SYS_BEGTEXT and a SYS_TEXTSTYLE have already been done
      // inputs:
      //        R14 = string buffer address in memblk[R95]
      //        R95 = memblk for string buffer
      // outputs:
      //        R39 = width
      //        R38 = height
      //        R37 = ascent

      Log.info("SYS_GETTEXTSIZE");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // get the string at that memory location
      String s = Mem.getString(theMem, offset);
      
      // getTextSize sets the width height into given Dimension, returns ascent
      Dimension d = new Dimension();
      int ascent = UI.getTextSize(s, d);

      Mem.SETREGW(regs, Mem.REG_39, (short)d.getWidth());
      Mem.SETREGW(regs, Mem.REG_38, (short)d.getHeight());
      Mem.SETREGW(regs, Mem.REG_37, (short)ascent);

      return;
    }


    case SYS_GETEOL: {
      // inputs:
      //        R15 = address in memblk R0 of where to push the OS's EOL seq
      //              On MS-Windows, this pushes out a 13 and a 10.
      //              On Linux and other OS's, it probably will push out ONLY 10

      Log.info("SYS_GETEOL");

      // put each char (as byte) at r15 onwards

      // get r15 offet
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // put newline separator at memory location, remembering new offset
      int newoffset = Mem.setString(kode, offset, newlineSep, false);
      Log.fine("new offset", newoffset);
      Mem.SETREGD(regs, Mem.REG_15, newoffset);

      return;
    }




    //*************************************
    // Music Functions
    //*************************************


    case SYS_PLAYSOUND: {
      // Inputs:
      //        R15 = offset to memblock[R95] memory containing filename of sound to play
      //        R95 = 0 if R0, 1 if R1, 2 if R2

      Log.info("SYS_PLAYSOUND");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // play that sound
      Music.playSound(Mem.getString(theMem, offset));

      return;
    }


    case SYS_PLAYMUSIC: {
      // Inputs:
      //        R15 = offset to memblock[R95] memory containing filename of music to play
      //        R95 = 0 if R0, 1 if R1, 2 if R2

      Log.info("SYS_PLAYMUSIC");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // play that music
      Music.playMusic(Mem.getString(theMem, offset));

      return;
    }


    case SYS_STOPMUSIC: {

      Log.info("SYS_STOPMUSIC");

      Music.stopMusic();

      return;
    }


    case SYS_ISMUSIC: {
      // outputs:
      //        R39=0 if no music playing, !=0 if music IS playing

      Log.info("SYS_ISMUSIC");

      Mem.SETREGW(regs, Mem.REG_39, Music.isMusicPlaying());

      return;
    }


    //****************************************
    // Print Functions
    //****************************************

    case SYS_PRINTSTART: {
      // Inputs:
      // Outputs:
      //        R15 = 0 if fail, otherwise non-0
      //        R14 = page size in X direction (total pixels in printable area)
      //        R13 = page size in Y direction (total pixels in printable area)
      //        R12 = printer resolution in X direction (pixels per inch)
      //        R11 = printer resolution in Y direction (pixels per inch)

      Log.info("SYS_PRINTSTART");

      // start printing, reporting if successful
      boolean printStarted = Printing.startPrinting();
      Mem.SETREGD(regs, Mem.REG_15, printStarted ? 1 : 0);

      if (printStarted) {
        // get page size
        Dimension d = Printing.getPageSize();
        int w = (int)d.getWidth();
        int h = (int)d.getHeight();
        Log.fine("page width heigth", w, h);
        Mem.SETREGD(regs, Mem.REG_14, w);
        Mem.SETREGD(regs, Mem.REG_13, h);

        // get resolution
        int r = Printing.getResolution();
        Log.fine("resolution", r);
        Mem.SETREGD(regs, Mem.REG_12, r);
        Mem.SETREGD(regs, Mem.REG_11, r);
      }

      return;
    }


    case SYS_PRINTBEGPAGE: {
      // Output:
      //        R95 = 1 if successful, 0 if fail

      Log.info("SYS_PRINTBEGPAGE");

      //        if( StartPage(PRThdc) > 0 ) SETREGB(REG_95, 1);
      //        else SETREGB(REG_95, 0);

      // begin next page, reporting if successful
      boolean worked = Printing.beginPage();
      Mem.SETREGB(regs, Mem.REG_95, worked ? (byte)1 : (byte)0);

      return;
    }


    case SYS_PRINTENDPAGE: {
      // Output:
      //        R95 = 1 if successful, 0 if fail

      Log.info("SYS_PRINTENDPAGE");

      //                if( EndPage(PRThdc) > 0 ) SETREGB(REG_95, 1);
      //        else SETREGB(REG_95, 0);

      // begin next page, reporting if successful
      boolean worked = Printing.endPage();
      Mem.SETREGB(regs, Mem.REG_95, worked ? (byte)1 : (byte)0);

      return;
    }


    case SYS_PRINTSTOP: {
      Log.info("SYS_PRINTSTOP");

      // just do it
      Printing.stopPrinting();

      return;
    }


    case SYS_PRINTTEXT: {
      // inputs:
      //        R14 = string buffer address in memblk[R95]
      //        R95 = memblk for string buffer
      //        R13 = textcolor
      //        R12 = x-coordinate of point
      //        R11 = y-coordinate of point
      //        R37 = font handle

      Log.info("SYS_PRINTTEXT");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // get the string at that memory location
      String s = Mem.getString(theMem, offset);

      int c = Mem.GETREGD(regs, Mem.REG_13);
      int x = Mem.GETREGD(regs, Mem.REG_12);
      int y = Mem.GETREGD(regs, Mem.REG_11);
      
      short fontHandle = Mem.GETREGW(regs, Mem.REG_37);

      Printing.printText(s, c, x, y, fontHandle);

      return;
    }


    case SYS_PRINTBM: {
      // inputs:
      //        R14 = handle of source bitmap
      //        R19 = x-coord of destination
      //        R18 = y-coord of destination
      //        R17 = x-coord of source
      //        R16 = y-coord of source
      //        R13 = width of source
      //        R12 = height of source
      //        R11 = width of destination
      //        R10 = height of destination

      Log.info("SYS_PRINTBM");

      int handle = Mem.GETREGD(regs, Mem.REG_14);
      int xDst = Mem.GETREGD(regs, Mem.REG_19);
      int yDst = Mem.GETREGD(regs, Mem.REG_18);
      int xSrc = Mem.GETREGD(regs, Mem.REG_17);
      int ySrc = Mem.GETREGD(regs, Mem.REG_16);
      int wSrc = Mem.GETREGD(regs, Mem.REG_13);
      int hSrc = Mem.GETREGD(regs, Mem.REG_12);
      int wDst = Mem.GETREGD(regs, Mem.REG_11);
      int hDst = Mem.GETREGD(regs, Mem.REG_10);

      Printing.printBitmap(handle, xDst, yDst, wDst, hDst,
                           xSrc, ySrc, wSrc, hSrc);

      return;
    }


    case SYS_PRINTLINES: {
      // inputs:
      //        R14 = offset in memblock[R95] of (x,y) DWORD coordinate array
      //        R95 = 0 for R0, 1 for R1, 2 for R2 memory block for R14 offset
      //        R13 = RGB color of lines
      //        R39 = # of points in the [R14] array

      Log.info("SYS_PRINTLINES");

      int colour = Mem.GETREGD(regs, Mem.REG_13);
      int nPoints = Mem.GETREGW(regs, Mem.REG_39);

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // create points array - for Java two arrays, one for X and one for Y
      Log.fine("creating points arrays, size", nPoints);
      int[] xs = new int[nPoints];
      int[] ys = new int[nPoints];

      // populate - entries a DWORDs
      for(int i=0; i<nPoints; i++) {
        xs[i] = Mem.GETMEMD(theMem, offset);
        offset = offset+4;
        ys[i] = Mem.GETMEMD(theMem, offset);
        offset = offset+4;
      }

      Printing.drawPolyline(colour, xs, ys, nPoints);

      return;
    }


    case SYS_PRINTGETTEXTSIZE: {
      // assumes a SYS_PRINTSTART and SYS_BEGTEXT have already been done
      // inputs:
      //        R14 = string buffer address in memblk[R95]
      //        R95 = memblk for string buffer
      //        R37 = font handle
      // outputs:
      //        R15 = width
      //        R14 = height
      //        R13 = height of non-descending portion of font

      Log.info("SYS_PRINTGETTEXTSIZE");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // get the string at that memory location
      String s = Mem.getString(theMem, offset);

      // getTextSize sets the width height into given Dimension, returns ascent
      Dimension d = new Dimension();
      int ascent = Printing.getTextSize(s, d, Mem.GETREGW(regs, Mem.REG_37));

      Mem.SETREGD(regs, Mem.REG_15, (int)d.getWidth());
      Mem.SETREGD(regs, Mem.REG_14, (int)d.getHeight());
      Mem.SETREGD(regs, Mem.REG_13, ascent);

      return;
    }



    //*********************************************
    // Memory Allocation Functions
    //*********************************************

    case SYS_ALLOCMEM: {
      // Inputs:
      //        R15 = size of memory block to allocate
      // Outputs:
      //        R15 = "handle" of memory block allocated (must be non-zero for 
      //        valid memory allocation)  This "handle" must be place in R1 or 
      //        R2 to access the memory (using LD1/LD2 type instructions)
      //        ==NULL if error
      // WARN: PTR4

      Log.info("SYS_ALLOCMEM");

      // find the size required
      long size = Mem.dUnsigned(Mem.GETREGD(regs, Mem.REG_15));

      // allocate memory, setting reg 15 to memory handle
      // we loose precision here, but who cares, memory would be too big anyway
      Mem.SETREGD(regs, Mem.REG_15, Mem.allocateMem((int)size));

      return;
    }


    case SYS_FREEMEM: {
      // Inputs:
      //        R15 = "handle" of memory block to free

      Log.info("SYS_FREEMEM");

      // find the memory handle
      int memoryHandle = Mem.GETREGD(regs, Mem.REG_15);

      // free the memory
      Mem.freeMem(memoryHandle);

      return;
    }



    //****************************************
    // Memory Manipulation Functions
    //****************************************

    case SYS_MEMSET: {
      // Inputs:
      //        R15 = offset in memblk[R95] of destination buffer
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      //        R94 = value to write into bytes of buffer
      //        R13 = number of bytes to set

      Log.info("SYS_MEMSET");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // get the byte value
      byte bVal = Mem.GETREGB(regs, Mem.REG_94);
      Log.fine("bVal", bVal);

      // get the count
      int count = Mem.GETREGD(regs, Mem.REG_13);
      Log.fine("count", count);

      // set the memory
      Mem.memset(theMem, offset, bVal, count);

      return;
    }


    case SYS_MEMCPY: {
      // Inputs:
      //        R15 = offset in memblk[R95] of destination buffer
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      //        R14 = offset in memblk[R94] of source buffer
      //        R94 = 0 for R0, 1 for R1, 2 for R2
      //        R13 = number of bytes to copy

      Log.info("SYS_MEMCPY");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get memory based on r94
      byte[] theMem2 = Mem.getMemByRegIndicator(regs, Mem.REG_94, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      int offset2 = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset2", offset2);

      // get the count
      int count = Mem.GETREGD(regs, Mem.REG_13);
      Log.fine("count", count);

      // copy second to first for count
      Mem.memcpy(theMem, offset, theMem2, offset2, count);

      return;
    }


    case SYS_MEMMOVE: {
      // Inputs:
      //        R15 = offset in memblk[R95] of destination buffer
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      //        R14 = offset in memblk[R94] of source buffer
      //        R94 = 0 for R0, 1 for R1, 2 for R2
      //        R13 = number of bytes to move

      Log.info("SYS_MEMCPY");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get memory based on r94
      byte[] theMem2 = Mem.getMemByRegIndicator(regs, Mem.REG_94, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      int offset2 = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset2", offset2);

      // get the count
      int count = Mem.GETREGD(regs, Mem.REG_13);
      Log.fine("count", count);

      // move second to first for count
      Mem.memmove(theMem, offset, theMem2, offset2, count);

      return;
    }


    case SYS_MEMCMP: {
      // Inputs:
      //        R15 = offset in memblk[R95] of buffer1
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      //        R14 = offset in memblk[R94] of buffer2
      //        R94 = 0 for R0, 1 for R1, 2 for R2
      //        R13 = number of bytes to compare
      // Outputs:
      //        R95 = <0 if buffer1<buffer2, 
      //               0 if buffer1==buffer2, 
      //              >0 if buffer1>buffer2
      //        E-flag set if R95==0, L-flag set if R95<0, G-flag set if R95>0

      Log.info("SYS_MEMCMP");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get memory based on r94
      byte[] theMem2 = Mem.getMemByRegIndicator(regs, Mem.REG_94, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      int offset2 = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset2", offset2);

      // get the count
      int count = Mem.GETREGD(regs, Mem.REG_13);
      Log.fine("count", count);

      int memcmp_res = Mem.memcmp(theMem, offset, theMem2, offset2, count);
      if (memcmp_res == 0) {
        // they are equal
        Mem.SETREGB(regs, Mem.REG_95, (byte)0);
      }
      else if (memcmp_res < 0) {
        // set to -1
        Mem.SETREGB(regs, Mem.REG_95, (byte)0xFF);
      }
      else {
        // set to 1
        Mem.SETREGB(regs, Mem.REG_95, (byte)0x01);
      }

      return;
    }



    //****************************************
    // String Functions
    //****************************************

    case SYS_STRCPY: {
      // Inputs:
      //        R15 = offset in memblk[R95] of destination buffer
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      //        R14 = offset in memblk[R94] of source buffer
      //        R94 = 0 for R0, 1 for R1, 2 for R2

      Log.info("SYS_STRCPY");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get memory based on r94
      byte[] theMem2 = Mem.getMemByRegIndicator(regs, Mem.REG_94, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      int offset2 = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset2", offset2);

      // copy the second string to the first
      Mem.strcpy(theMem, offset, theMem2, offset2);

      return;
    }


    case SYS_STRLEN: {
      // Inputs:
      //        R15 = offset in memblk[R95] of buffer
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      // Outputs:
      //        R39 = length of nul-terminated string in buffer

      Log.info("SYS_STRLEN");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      Mem.SETREGW(regs, Mem.REG_39, (short)Mem.strlen(theMem, offset));

      return;
    }


    case SYS_STRCMP: {
      // Inputs:
      //      R15 = offset in memblk[R95] of buffer1
      //      R95 = 0 for R0, 1 for R1, 2 for R2
      //      R14 = offset in memblk[R94] of buffer2
      //      R94 = 0 for R0, 1 for R1, 2 for R2
      // Outputs:
      //      R95 = <0 if buffer1<buffer2, 
      //             0 if buffer1==buffer2, 
      //            >0 if buffer1>buffer2

      Log.info("SYS_STRCMP");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get memory based on r94
      byte[] theMem2 = Mem.getMemByRegIndicator(regs, Mem.REG_94, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      int offset2 = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset2", offset2);

      Mem.SETREGB(regs, Mem.REG_95, (byte)Mem.strcmp(theMem, offset, theMem2, offset2));

      return;
    }


    case SYS_STRICMP: {
      // Inputs:
      //      R15 = offset in memblk[R95] of buffer1
      //      R95 = 0 for R0, 1 for R1, 2 for R2
      //      R14 = offset in memblk[R94] of buffer2
      //      R94 = 0 for R0, 1 for R1, 2 for R2
      // Outputs:
      //      R95 = <0 if buffer1<buffer2, 
      //             0 if buffer1==buffer2, 
      //            >0 if buffer1>buffer2

      Log.info("SYS_STRICMP");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get memory based on r94
      byte[] theMem2 = Mem.getMemByRegIndicator(regs, Mem.REG_94, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      int offset2 = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset2", offset2);

      Mem.SETREGB(regs, Mem.REG_95, (byte)Mem.stricmp(theMem, offset, theMem2, offset2));

      return;
    }



    //****************************************
    // File I/O Functions
    //****************************************

    case SYS_FILESIZE: {
      // inputs:
      //        R15 = offset of filename in memblk [R95]
      //        R95 = memblk of filename
      //        R94 = folder (see SYS_FOPEN)
      // outputs:
      //        R13 = size of file (-1 if error)

      Log.info("SYS_FILESIZE");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // get the filename
      String filename = FileIO.buildPath(Mem.getString(theMem, offset),
                                         Mem.GETREGB(regs, Mem.REG_94), false);

      // put the file size in R13 (-1 on failure)
      Mem.SETREGD(regs, Mem.REG_13, FileIO.size(filename));

      return;
    }


    case SYS_FOPEN: {
      // inputs:
      //        R15 = offset of filename in memory block specified by R95
      //        R95 = memory block # of filename (0 if R0 based, 1 if R1 based, 2 if R2 based)
      //        R94 = folder, as in:
      //          0 = folder the game is installed in (home)
      //          1 = UI (folder containing UI image files)
      //          2 = IMAGES (folder containing game image set files)
      //          3 = BACKGNDS ( folder containing 'decorative' bitmap files)
      //          4 = MUSIC (folder containing music files)
      //          5 = SOUNDS (folder containing sound effect files)
      //
      // In the Windows implementation, UI, IMAGES, and BACKGNDS files
      // are all .BMP files, and get that file extendsion added to them
      // automatically.  MUSIC files are assumed to be MIDI files and get .MID.
      // SOUNDS are assumed to be .WAV files, and get .WAV appended.
      // When porting this to another platform (or even under Windows),
      // there's absolutely NO reason why you can't use other graphics file
      // formats (so long as they're 'lossless', so as to preserve the exact
      // colors of the pixels, so that transparency works right), or other
      // music (like MP3) or sound file formats.  As long as you modify
      // the KINT code that opens and loads/plays the files, you can use
      // whatever file formats you want.  The KASM code doesn't care.
      //
      //      R39 = open flags.  Bits are:
      //               _O_RDONLY      0x0000  /* open for reading only */
      //               _O_WRONLY      0x0001  /* open for writing only */
      //               _O_RDWR        0x0002  /* open for reading and writing */
      //               _O_APPEND      0x0008  /* writes done at eof */
      //               _O_CREAT       0x0100  /* create and open file */
      //               _O_TRUNC       0x0200  /* open and truncate */
      //               _O_TEXT        0x4000  /* file mode is text (translated) */
      //               _O_BINARY      0x8000  /* file mode is binary (untranslated) */
      // outputs:
      //      R15 = file handle if successful, -1 if failed

      Log.info("SYS_FOPEN");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // get the filename
      String filename = FileIO.buildPath(Mem.getString(theMem, offset),
                                         Mem.GETREGB(regs, Mem.REG_94), false);

      // open the file and put file handle in r15
      Mem.SETREGD(regs, Mem.REG_15, FileIO.open(filename, Mem.GETREGW(regs, Mem.REG_39)));

      return;
    }


    case SYS_FEOF: {
      // Inputs:
      //        R15 = file handle
      // Outputs:
      //        R12 = 1 if at EOF, 0 if not at EOF, -1 if error

      Log.info("SYS_FEOF");

      // get the file handle
      int fileHandle = Mem.GETREGD(regs, Mem.REG_15); 

      Mem.SETREGD(regs, Mem.REG_12, FileIO.eof(fileHandle));

      return;
    }


    case SYS_FREAD: {
      // Inputs:
      //        R15 = file handle
      //        R14 = buffer offset
      //        R95 = memory block # (0 R0 based, 1 R1 based, 2 R2 based)
      //        R13 = max number of bytes to read
      // Outputs:
      //        R12 = number of bytes read.  (-1 if error)

      Log.info("SYS_FREAD");

      // get the file handle
      int fileHandle = Mem.GETREGD(regs, Mem.REG_15); 

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // get the count
      int count = Mem.GETREGD(regs, Mem.REG_13);

      // do the read and put result in r12
      Mem.SETREGD(regs, Mem.REG_12, FileIO.read(fileHandle, theMem, offset, 
                                                count));

      return;
    }


    case SYS_FWRITE: {
      // Inputs:
      //        R15 = file handle
      //        R14 = buffer offset in memory block specified by R95
      //        R13 = number of bytes to write
      //        R95 = memory block # (0 R0 based, 1 R1 based, 2 R2 based)
      // Outputs:
      //        R12 = number of bytes written.  (-1 if error)


      // get the file handle
      int fileHandle = Mem.GETREGD(regs, Mem.REG_15); 

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offset
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      Log.fine("offset", offset);

      // get the count
      int count = Mem.GETREGD(regs, Mem.REG_13);

      // do the write and put result in r12
      Mem.SETREGD(regs, Mem.REG_12, FileIO.write(fileHandle, theMem, offset, count));

      return;
    }


    case SYS_FSEEK: {
      // Inputs:
      //        R15 = file handle
      //        R14 = offset to move the file pointer to
      //        R95 = origin:
      //                SEEK_SET    0
      //                SEEK_CUR    1
      //                SEEK_END    2
      // Outputs:
      //        R12 = offset in bytes from the beginning of the file of 
      //              the new position (-1 if error)

      Log.info("SYS_FSEEK");

      int fileHandle = Mem.GETREGD(regs, Mem.REG_15);
      int offset = Mem.GETREGD(regs, Mem.REG_14);
      int origin = Mem.GETREGB(regs, Mem.REG_95);

      Mem.SETREGD(regs, Mem.REG_12, FileIO.seek(fileHandle, offset, origin));

      return;
    }


    case SYS_FCLOSE: {
      // Inputs:
      //        R15 = file handle to close

      Log.info("SYS_FCLOSE");

      FileIO.close(Mem.GETREGD(regs, Mem.REG_15));

      return;
    }


    case SYS_FFINDFIRST: {
      // Inputs:
      //        R15 = offset in memblock[R95] of filespec to search for
      //        R95 = 0 for R0, 1 for R1, 2 for R2
      //        R14 = offset in memblock[R94] of buffer for found filename 
      //              (must be minimum of 260 bytes)
      //        R94 = 0 for R0, 1 for R1, 2 for R2
      //        R93 = folder, as in:
      //              0 = folder the game is installed in (home)
      //              1 = UI (folder containing UI image files)
      //              2 = IMAGES (folder containing game image set files)
      //              3 = BACKGNDS (folder containing 'decorative' bitmap files)
      //              4 = MUSIC (folder containing music files)
      //              5 = SOUNDS (folder containing sound effect files)
      // Outputs:
      //        R15 = "findfirst/next" handle of first file found, or -1 if 
      //              none found or error
      //              and found file name in buffer [R94]:R14
      //        R95 = file attributes: 0x00 = _A_NORMAL, 0x10 = _A_SUBDIR
      // NOTE: the returned file name must be JUST the filename, not
      //        including any part of the full path or the file extension.

      Log.info("SYS_FFINDFIRST");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // get the folder type
      byte folderType = Mem.GETREGB(regs, Mem.REG_93);
      Log.fine("folderType", folderType);

      // get the filespec, based on R93 folder
      String filespec = FileIO.buildPath(Mem.getString(theMem, offset),
                                         folderType, false);

      // find the first file of this spec, returns null if none
      FindFirstFiles foundFiles = FileIO.findFirst(filespec);

      if (foundFiles == null) {
        // no matching file
        Mem.SETREGD(regs, Mem.REG_15, -1);      
      }
      else {
        // set the handle for this findfirst result
        Mem.SETREGD(regs, Mem.REG_15, foundFiles.getHandle());

        // complete the call
        findFirstNextCommon(foundFiles, folderType, regs, kode);
      }

      return;
    }


    case SYS_FFINDNEXT: {
      // Inputs:
      //        R15 = "find first/next" handle returned by previous FINDFIRST
      //        R14 = offset in memblock[R94] of buffer for found filename 
      //              (must be minimum of 260 bytes)
      //        R94 = 0 for R0, 1 for R1, 2 for R2
      // Outputs:
      //        R15 = "findfirst/next" handle of file found [actually 0]
      //                               or -1 if none found or error
      //              and found file name in buffer [R94]:R14
      //  R95 = file attributes: 0x00 = _A_NORMAL, 0x10 = _A_SUBDIR

      Log.info("SYS_FFINDNEXT");

      // make the call, returns null if no more
      FindFirstFiles foundFiles = FileIO.findNext(Mem.GETREGD(regs, Mem.REG_15));

      if (foundFiles == null) {
        // no more matching files
        Mem.SETREGD(regs, Mem.REG_15, -1);      
      }
      else {
        // okay, another file found
        Mem.SETREGD(regs, Mem.REG_15, 0);

        // get the folder type
        byte folderType = Mem.GETREGB(regs, Mem.REG_93);
        Log.fine("folderType", folderType);

        // complete the call
        findFirstNextCommon(foundFiles, folderType, regs, kode);
      }

      return;
    }


    case SYS_FFINDCLOSE: {
      // Inputs:
      //      R15 = "find first/next" handle returned by previous FINDFIRST

      Log.info("SYS_FFINDCLOSE");

      // make the call, don't care about the result
      FileIO.findClose(Mem.GETREGD(regs, Mem.REG_15));

      return;
    }


    case SYS_DELETE: {
      // Inputs:
      //      R15 = address of filename
      //      R95 = memblk of filename

      Log.info("SYS_DELETE");

      // get memory based on r95
      byte[] theMem = Mem.getMemByRegIndicator(regs, Mem.REG_95, kode);

      // get the offsets
      int offset = Mem.GETREGD(regs, Mem.REG_15);
      Log.fine("offset", offset);

      // get the filename
      String filename = FileIO.buildPath(Mem.getString(theMem, offset),
                                         FileIO.FOLDER_HOME, false);

      // just delete it, no return value required
      FileIO.delete(filename);

      return;
    }


    case SYS_MAKESHARECOPY: {
      // Inputs:
      //        R95 = command
      //        R15 = address of buffer in memblk R0 for next filename
      //        R13 = address of buffer in memblk R0 containing destination path 
      //              (if R95 != 0)
      // Outputs:
      //        R14 = -1 if error
      //              else total size of files to be copied if cmd==0 
      //              (and filename of FIRST file in R15 buffer)
      //              else 0 (and filename of NEXT file in R15 buffer, or NUL 
      //              if last file)
      //
      // NOTE: KASM code always calls this FIRST with command=0 (read file 
      // names, return first file name in buffer R15 and total size in R14)
      // followed by a sequential series of calls with R95 != 0 to copy 
      // successive files.  Each successive call returns the NEXT filename to 
      // be copied in buffer R15.
      // When no more files to be copied, buffer R15 starts with a NUL.
      //
      // NOTE: this function is FAR more complex than ordinarily need be, but I 
      // didn't want the KASM code to know anything about the files being copied
      // for the shareware version (as their names and/or formats will 
      // undoubtedly be different on different OS's), and I didn't want the KINT
      // interpreter (this code) to know anything about the specific list of 
      // files being copied either, as I want the KINT code to remain generic 
      // across all of the games.  So, the files to be copied are in an ASCII 
      // file call "swlist", as a comma-separated list of src/dst pairs.  
      // For example:
      //        eks0.ebk,setup.exe,eksctlss.ebk,setup.ctl,ekssubk.ebk,setupbk.ebk,shs.epk,shs.epk
      // all on one line, no CRs, no LFs, source filename first, comma, 
      // destination filename second, comma, next pair....etc.

      Log.info("SYS_MAKESHARECOPY");

      Mem.SETREGD(regs, Mem.REG_14, -1);

      return;
    }


    default:
      Log.errorExit("Unknown system function", sysfunc);
    }
  }



  // common code for the result of findFirst and findNext
  private static void findFirstNextCommon(FindFirstFiles foundFiles, 
                                          byte folderType,
                                          byte[] regs, byte[] kode) {
    // set the file attributes for the file we are on
    Mem.SETREGB(regs, Mem.REG_95, foundFiles.getCurrentAttributes());

    // set the file name in the given memory

    // get memory based on r94 to store the result
    byte[] resultMem = Mem.getMemByRegIndicator(regs, Mem.REG_94, kode);

    // and the offset in that result mem
    int resultOffset = Mem.GETREGD(regs, Mem.REG_14);
    Log.fine("resultOffset", resultOffset);

    // we need to drop the file extension from the name, for all but folderType 0
    String foundName = foundFiles.getCurrentName();
    if (folderType != 0) {
      foundName = FileIO.dropFilenameExt(foundName);
    }
    Mem.setString(resultMem, resultOffset, foundName, true);
  }




  // Return our CVS id
  public static String getCVSID() {
    return "$Id: SysFuncs.java,v 1.6 2002/11/28 19:32:28 Bennett Stephen Exp $";
  }



}


/*
 * $Log: SysFuncs.java,v $
 * Revision 1.6  2002/11/28 19:32:28  Bennett Stephen
 * release candidate 1 version
 *
 * Revision 1.5  2002/10/12 14:22:15  Bennett Stephen
 * printing and music added
 *
 * Revision 1.4  2002/10/06 15:24:58  Bennett Stephen
 * safe version, still being worked on
 *
 * Revision 1.3  2002/09/08 19:39:27  Bennett Stephen
 * a lot more sys funcs added
 *
 * Revision 1.2  2002/08/10 08:20:57  Bennett Stephen
 * mem funcs added and tidy up
 *
 * Revision 1.1  2002/08/07 19:22:19  Bennett Stephen
 * Initial Version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
