// A Java version for the Kaser Interpreter - General Utilities
// For some of my own fun!
// Stephen M. Bennett
// August 2002

package stephen.bennett.kint;


// the imports
import java.util.*;


// here we go
public class Utils {

  // **********************************************
  // Handle <-> Vector Utils
  // **********************************************

  // add an object to given handles, re-use empty indexes,
  // returning handle it was placed at
  public static int addObjectToHandles(Object obj, Vector handles, 
                                       boolean firstIndexIsOne, String msg2) {
    // find next place to put it
    for(int i=0; i<handles.size(); i++) {
      if (handles.elementAt(i) == null) {
        // this is a free space - use it
        handles.setElementAt(obj, i);

        // define handle value
        int handleValue = computeHandleFromIndex(i, firstIndexIsOne);

        Log.fine("reused handle", msg2, handleValue);
        return handleValue;
      }
    }

    // no free entries, put at end
    handles.add(obj);
    int handleValue = computeHandleFromIndex(handles.size()-1, firstIndexIsOne);
    Log.fine("new handle", msg2, handleValue);
    return handleValue;
  }


  // get the object given the handle
  public static Object getObjectByHandle(int handle, Vector handles, 
                                         boolean firstIndexIsOne, String msg2) {
    Log.fine("getting object at handle", msg2, handle);

    // compute the index
    int i = computeIndexFromHandle(handle, firstIndexIsOne);

    // could do some bounds checking here
    if (i < 0 || i >= handles.size()) {
      Log.errorExit("Handle "+handle+" (providing index "+i+") is out of bounds for "+msg2);
    }

    return handles.elementAt(i);
  }


  // remove the object at the given handle
  public static void removeObjectByHandle(int handle, Vector handles, 
                                          boolean firstIndexIsOne, String msg2) {
    Log.fine("removing object at handle", msg2, handle);

    // compute the index
    int i = computeIndexFromHandle(handle, firstIndexIsOne);

    // could do some bounds checking here

    // remove that object
    handles.setElementAt(null, i);
  }


  // support method for handles, to decide if first handle is 0 or 1
  private static int computeHandleFromIndex(int i, boolean firstIndexIsOne) {
    if (firstIndexIsOne) {
      return i+1;
    }
    
    return i;
  }

  // support method for handles, to decide if first handle is 0 or 1
  private static int computeIndexFromHandle(int handle, 
                                            boolean firstIndexIsOne) {
    if (firstIndexIsOne) {
      return handle-1;
    }
    
    return handle;
  }




  // Return our CVS id
  public static String getCVSID() {
    return "$Id: Utils.java,v 1.2 2002/10/06 15:23:17 Bennett Stephen Exp $";
  }


}


/*
 * $Log: Utils.java,v $
 * Revision 1.2  2002/10/06 15:23:17  Bennett Stephen
 * CVS method, and handle index check
 *
 * Revision 1.1  2002/09/08 19:36:00  Bennett Stephen
 * intial version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
